from __future__ import print_function
from builtins import (range, str)

import datetime
import tempfile
import os
import unittest
import atexit
import inspect
import subprocess
import numpy

import cfdm

warnings = False

tmpfile  = tempfile.mktemp('.cfdm_test')
tmpfileh  = tempfile.mktemp('.cfdm_test')
tmpfilec  = tempfile.mktemp('.cfdm_test')
tmpfiles = [tmpfile, tmpfileh, tmpfilec]
def _remove_tmpfiles():
    '''
'''
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass

        
atexit.register(_remove_tmpfiles)


class read_writeTest(unittest.TestCase):
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'test_file.nc')

    test_only = []
#    test_only = ['NOTHING!!!!!']
#    test_only = ['test_read_CDL']
#    test_only = ['test_write_HDF_chunks']
#    test_only = ['test_read_write_unlimited']
#    test_only = ['test_read_field']
#    test_only = ['test_write_datatype']
    
    def test_read_field(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        # Test field keyword of cfdm.read
        filename = self.filename
        
        f = cfdm.read(filename)
        self.assertTrue(len(f) == 1, '\n'+str(f))

        f = cfdm.read(filename, extra=['dimension_coordinate'], warnings=warnings)
        self.assertTrue(len(f) == 4, '\n'+str(f))

        f = cfdm.read(filename, extra=['auxiliary_coordinate'], warnings=warnings)
        self.assertTrue(len(f) == 4, '\n'+str(f))
        
        f = cfdm.read(filename, extra='cell_measure')
        self.assertTrue(len(f) == 2, '\n'+str(f))

        f = cfdm.read(filename, extra=['field_ancillary'])
        self.assertTrue(len(f) == 4, '\n'+str(f))
                
        f = cfdm.read(filename, extra='domain_ancillary', warnings=warnings)
        self.assertTrue(len(f) == 4, '\n'+str(f))
        
        f = cfdm.read(filename, extra=['field_ancillary', 'auxiliary_coordinate'],
                      warnings=warnings)
        self.assertTrue(len(f) == 7, '\n'+str(f))
        
        self.assertTrue(len(cfdm.read(filename, extra=['domain_ancillary', 'auxiliary_coordinate'], warnings=warnings)) == 7)
        self.assertTrue(len(cfdm.read(filename, extra=['domain_ancillary', 'cell_measure', 'auxiliary_coordinate'], warnings=warnings)) == 8)

        f = cfdm.read(filename, extra=('field_ancillary', 'dimension_coordinate',
                                       'cell_measure', 'auxiliary_coordinate',
                                       'domain_ancillary'), warnings=warnings)
        self.assertTrue(len(f) == 14, '\n'+str(f))


    def test_read_write_format(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cfdm.read(self.filename)[0]
        for fmt in ('NETCDF3_CLASSIC',
                    'NETCDF3_64BIT',
                    'NETCDF4',
                    'NETCDF4_CLASSIC'):
            cfdm.write(f, tmpfile, fmt=fmt)
            g = cfdm.read(tmpfile)[0]
            self.assertTrue(f.equals(g, verbose=True),
                            'Bad read/write of format: {}'.format(fmt))


    def test_read_write_netCDF4_compress_shuffle(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cfdm.read(self.filename)[0]
        for fmt in ('NETCDF4',
                    'NETCDF4_CLASSIC'):
            for shuffle in (True, False):
                for compress in range(10):
                    cfdm.write(f, tmpfile, fmt=fmt,
                               compress=compress,
                               shuffle=shuffle)
                    g = cfdm.read(tmpfile)[0]
                    self.assertTrue(
                        f.equals(g, verbose=True),
                        'Bad read/write with lossless compression: {}, {}, {}'.format(fmt, compress, shuffle))
        #--- End: for


    def test_read_write_missing_data(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cfdm.read(self.filename)[0]
        for fmt in ('NETCDF3_CLASSIC',
                    'NETCDF3_64BIT',
                    'NETCDF4',
                    'NETCDF4_CLASSIC'):
            cfdm.write(f, tmpfile, fmt=fmt)
            g = cfdm.read(tmpfile)[0]
            self.assertTrue(f.equals(g, verbose=True),
                            'Bad read/write of format: {}'.format(fmt))


    def test_write_datatype(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cfdm.read(self.filename)[0] 
        self.assertTrue(f.data.dtype == numpy.dtype(float))

        cfdm.write(f, tmpfile, fmt='NETCDF4', 
                 datatype={numpy.dtype(float): numpy.dtype('float32')})
        g = cfdm.read(tmpfile)[0]
        self.assertTrue(g.data.dtype == numpy.dtype('float32'), 
                        'datatype read in is '+str(g.data.dtype))


    def test_read_write_unlimited(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        for fmt in ('NETCDF4', 'NETCDF3_CLASSIC'):
            f = cfdm.read(self.filename)[0]
            
            f.domain_axes['domainaxis0'].nc_set_unlimited(True)
            cfdm.write(f, tmpfile, fmt=fmt)
            
            f = cfdm.read(tmpfile)[0]
            self.assertTrue(f.domain_axes['domainaxis0'].nc_is_unlimited())

        fmt = 'NETCDF4'
        f = cfdm.read(self.filename)[0]
        f.domain_axes['domainaxis0'].nc_set_unlimited(True)
        f.domain_axes['domainaxis2'].nc_set_unlimited(True)
        cfdm.write(f, tmpfile, fmt=fmt)
        
        f = cfdm.read(tmpfile)[0]
        self.assertTrue(f.domain_axes['domainaxis0'].nc_is_unlimited())
        self.assertTrue(f.domain_axes['domainaxis2'].nc_is_unlimited())


    def test_read_CDL(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        subprocess.run(' '.join(['ncdump', self.filename, '>', tmpfile]),
                       shell=True, check=True)
        subprocess.run(' '.join(['ncdump', '-h', self.filename, '>', tmpfileh]),
                                shell=True, check=True)
        subprocess.run(' '.join(['ncdump', '-c', self.filename, '>', tmpfilec]),
                       shell=True, check=True)

        f0 = cfdm.read(self.filename)[0]
        f = cfdm.read(tmpfile)[0]
        h = cfdm.read(tmpfileh)[0]
        c = cfdm.read(tmpfilec)[0]

        self.assertTrue(f0.equals(f, verbose=True))

        self.assertTrue(f.construct('grid_latitude').equals(c.construct('grid_latitude'), verbose=True))
        self.assertTrue(f0.construct('grid_latitude').equals(c.construct('grid_latitude'), verbose=True))

        with self.assertRaises(OSError):
            x = cfdm.read('test_read_write.py')
            
        
#--- End: class

if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    print(cfdm.environment(display=False))
    print('')
    unittest.main(verbosity=2)



from . import abstract


class InteriorRing(abstract.PropertiesData):
    '''An interior ring array with properties.

If a cell is composed of multiple polygon parts, an individual polygon
may define an "interior ring", i.e. a region that is to be omitted
from, as opposed to included in, the cell extent. In this case an
interior ring array is required that records whether each polygon is
to be included or excluded from the cell, and is supplied by an
interior ring variable in CF-netCDF. The interior ring array spans the
same domain axis constructs as its coordinate array, with the addition
of an extra dimension that indexes the parts for each cell. For
example, a cell describing the land area surrounding a lake would
require two polygon parts: one defines the outer boundary of the land
area; the other, recorded as an interior ring, is the lake boundary,
defining the inner boundary of the land area.

.. versionadded:: 1.8.0

    '''

#--- End: class



from __future__ import print_function
from builtins import (str, super)

from . import mixin
from . import core

from . import Constructs


class Domain(mixin.ConstructAccess,
             mixin.Container,
             core.Domain):
    '''A domain of the CF data model.

The domain represents a set of discrete "locations" in what generally
would be a multi-dimensional space, either in the real world or in a
model's simulated world. These locations correspond to individual data
array elements of a field construct

The domain is defined collectively by the following constructs of the
CF data model: domain axis, dimension coordinate, auxiliary
coordinate, cell measure, coordinate reference and domain ancillary
constructs.

.. versionadded:: 1.7.0

    '''
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._Constructs = Constructs
        return instance
    #--- End: def
    
    def __repr__(self):
        '''Called by the `repr` built-in function.

x.__repr__() <==> repr(x)

        '''
        shape = sorted([domain_axis.get_size()
                        for domain_axis in list(self.domain_axes.values())])
        shape = str(shape)
        shape = shape[1:-1]
        
        return '<{0}: {{{1}}}>'.format(self.__class__.__name__, shape)
    #--- End: def

    def __str__(self):
        '''Called by the `str` built-in function.

x.__str__() <==> str(x)

        '''
        def _print_item(self, cid, variable, axes):
            '''Private function called by __str__

            '''            
            x = [variable.identity(default='key%{0}'.format(cid))]

            if variable.has_data():
                shape = [axis_names[axis] for axis in axes]
                shape = str(tuple(shape)).replace("'", "")
                shape = shape.replace(',)', ')')
                x.append(shape)
            elif (variable.construct_type in ('auxiliary_coordinate', 'domain_ancillary') and
                  variable.has_bounds() and variable.bounds.has_data()):                
                # Construct has no data but it does have bounds data
                shape = [axis_names[axis] for axis in axes]
                shape.extend([str(n) for n in variable.bounds.data.shape[len(axes):]])
                shape = str(tuple(shape)).replace("'", "")
                shape = shape.replace(',)', ')')
                x.append(shape)
            elif hasattr(variable, 'nc_get_external'):
                if variable.nc_get_external():
                    ncvar = variable.nc_get_variable(None)
                    if ncvar is not None:
                        x.append(' (external variable: ncvar%{})'.format(ncvar))
                    else:
                        x.append(' (external variable)')
            #--- End: if
                
            if variable.has_data():
                x.append(' = {0}'.format(variable.data))
            elif (variable.construct_type in ('auxiliary_coordinate', 'domain_ancillary') and
                  variable.has_bounds() and variable.bounds.has_data()):
                # Construct has no data but it does have bounds data
                x.append(' = {0}'.format(variable.bounds.data))
               
            return ''.join(x)
        #--- End: def
                          
        string = []
        
        axis_names = self._unique_domain_axis_identities()

        constructs_data_axes = self.constructs.data_axes()
        
        x = []
        for axis_cid in sorted(self.domain_axes):
            for cid, dim in list(self.dimension_coordinates.items()):
                if constructs_data_axes[cid] == (axis_cid,):
                    name = dim.identity(default='key%{0}'.format(cid))
                    y = '{0}({1})'.format(name, dim.get_data().size)
                    if y != axis_names[axis_cid]:
                        y = '{0}({1})'.format(name, axis_names[axis_cid])
                    if dim.has_data():
                        y += ' = {0}'.format(dim.get_data())
                        
                    x.append(y)
        #--- End: for
        if x:
            string.append('Dimension coords: {}'.format('\n                : '.join(x)))

        # Auxiliary coordinates
        x = [_print_item(self, cid, v, constructs_data_axes[cid]) 
             for cid, v in sorted(self.auxiliary_coordinates.items())]
        if x:
            string.append('Auxiliary coords: {}'.format(
                '\n                : '.join(x)))
        
        # Cell measures
        x = [_print_item(self, cid, v, constructs_data_axes[cid])
             for cid, v in sorted(self.cell_measures.items())]
        if x:
            string.append('Cell measures   : {}'.format(
                '\n                : '.join(x)))
            
        # Coordinate references
        x = sorted([str(ref) for ref in list(self.coordinate_references.values())])
        if x:
            string.append('Coord references: {}'.format(
                '\n                : '.join(x)))
            
        # Domain ancillary variables
        x = [_print_item(self, cid, anc, constructs_data_axes[cid])
             for cid, anc in sorted(self.domain_ancillaries.items())]
        if x:
            string.append('Domain ancils   : {}'.format(
                '\n                : '.join(x)))
                                      
        string.append('')

        return '\n'.join(string)
    #--- End def
                          
    def _dump_axes(self, axis_names, display=True, _level=0):
        '''Return a string containing a description of the domain axes of the
field.
    
:Parameters:
    
    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.
    
    _level: `int`, optional

:Returns:
    
    `str`
        A string containing the description.
    
**Examples:**

        '''
        indent1 = '    ' * _level
        indent2 = '    ' * (_level+1)

        axes = self.domain_axes

        w = sorted(["{0}Domain Axis: {1}".format(indent1, axis_names[axis])
                    for axis in axes])

        string = '\n'.join(w)

        if display:
            print(string)
        else:
            return string
    #--- End: def
    
    def dump(self, display=True, _level=0, _title=None):
        '''A full description of the domain.

The domain components are described without abbreviation with the
exception of data arrays, which are abbreviated to their first and
last values.

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.

          *Parameter example:*
            ``f.dump()`` is equivalent to ``print
            f.dump(display=False)``.

:Returns:

    `None` or `str`
        If *display* is True then the description is printed and
        `None` is returned. Otherwise the description is returned as a
        string.

        '''
        indent = '    '      
        indent0 = indent * _level
        indent1 = indent0 + indent

        axis_to_name = self._unique_domain_axis_identities()

        construct_name = self._unique_construct_names()

        constructs_data_axes = self.constructs.data_axes()
        
        string = []

        # Domain axes
        axes = self._dump_axes(axis_to_name, display=False, _level=_level)
        if axes:
            string.append(axes)
          
        # Dimension coordinates
        for cid, value in sorted(self.dimension_coordinates.items()):
            string.append('')
            string.append(
                value.dump(display=False, _level=_level,
                           _title='Dimension coordinate: {0}'.format(
                               construct_name[cid]),
                           _axes=constructs_data_axes[cid],
                           _axis_names=axis_to_name))
            
        # Auxiliary coordinates
        for cid, value in sorted(self.auxiliary_coordinates.items()):
            string.append('')
            string.append(
                value.dump(display=False, _level=_level,
                           _title='Auxiliary coordinate: {0}'.format(
                               construct_name[cid]),
                           _axes=constructs_data_axes[cid],
                           _axis_names=axis_to_name))

        # Domain ancillaries
        for cid, value in sorted(self.domain_ancillaries.items()):
            string.append('') 
            string.append(value.dump(display=False, _level=_level,
                                     _title='Domain ancillary: {0}'.format(
                                         construct_name[cid]),
                                     _axes=constructs_data_axes[cid],
                                     _axis_names=axis_to_name))
            
        # Coordinate references
        for cid, value in sorted(self.coordinate_references.items()):
            string.append('')
            string.append(
                value.dump(display=False, _level=_level,
                           _title='Coordinate reference: {0}'.format(
                               construct_name[cid]),
                           _construct_names=construct_name,
                           _auxiliary_coordinates=tuple(self.auxiliary_coordinates),
                           _dimension_coordinates=tuple(self.dimension_coordinates)))
            
        # Cell measures
        for cid, value in sorted(self.cell_measures.items()):
            string.append('')
            string.append( value.dump(
                display=False, _key=cid,
                _level=_level, _title='Cell measure: {0}'.format(construct_name[cid]),
                _axes=constructs_data_axes[cid],
                _axis_names=axis_to_name))
            

        string.append('')
        
        string = '\n'.join(string)
       
        if display:
            print(string)
        else:
            return string
    #--- End: def

    def equals(self, other, rtol=None, atol=None, verbose=False,
               ignore_data_type=False, ignore_fill_value=False,
               ignore_compression=False, ignore_type=False):
        '''Whether two domains are the same.

.. versionadded:: 1.7.0

**Examples:**

>>> d.equals(d)
True
>>> d.equals(d.copy())
True
>>> d.equals('not a domain')
False
        '''
        pp = super()._equals_preprocess(other, verbose=verbose,
                                        ignore_type=ignore_type)
        if pp is True or pp is False:
            return pp
        
        other = pp

        # ------------------------------------------------------------
        # Check the constructs
        # ------------------------------------------------------------              
        if not self._equals(self.constructs, other.constructs,
                            rtol=rtol, atol=atol, verbose=verbose,
                            ignore_data_type=ignore_data_type,
                            ignore_fill_value=ignore_fill_value,
                            ignore_compression=ignore_compression):
            if verbose:
                print("{0}: Different metadata constructs".format(
                    self.__class__.__name__))
            return False

        return True
    #--- End: def
        
#--- End: class



from builtins import super

from . import abstract
from . import mixin

from . import Constructs


class Domain(mixin.ConstructAccess, abstract.Container):
    '''A domain of the CF data model.

The domain represents a set of discrete "locations" in what generally
would be a multi-dimensional space, either in the real world or in a
model's simulated world. These locations correspond to individual data
array elements of a field construct

The domain is defined collectively by the following constructs of the
CF data model: domain axis, dimension coordinate, auxiliary
coordinate, cell measure, coordinate reference and domain ancillary
constructs.

.. versionadded:: 1.7.0

    '''

    # Define the base of the identity keys for each construct type
    _construct_key_base = {'auxiliary_coordinate': 'auxiliarycoordinate',
                           'cell_measure'        : 'cellmeasure',
                           'coordinate_reference': 'coordinatereference',
                           'dimension_coordinate': 'dimensioncoordinate',
                           'domain_ancillary'    : 'domainancillary',
                           'domain_axis'         : 'domainaxis',
    }
    

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._Constructs = Constructs
        return instance
    #--- End: def
    
    def __init__(self, source=None, copy=True, _use_data=True):
        '''**Initialization**

:Parameters:

    source: optional
        Initialize the metadata constructs from those of *source*.
        
        A new domain may also be instantiated with the
        `fromconstructs` class method.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__()
        
        if source is not None:
            try:                
                constructs = source.constructs
            except AttributeError:
                constructs = self._Constructs(**self._construct_key_base)
                copy = False
                _use_data = True            
            else:
                constructs = constructs._view(ignore=('cell_method',
                                                      'field_ancillary'))
        else:
            constructs = self._Constructs(**self._construct_key_base)
            copy = False
            _use_data = True

        if copy or not _use_data:
            constructs = constructs.copy(data=_use_data)
            
        self._set_component('constructs', constructs, copy=False)
    #--- End: def
    
    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------   
    @property
    def constructs(self):
        '''Return the metdata constructs.

.. versionadded:: 1.7.0

:Returns:

    `Constructs`
        The constructs.

**Examples:**

>>> print(d.constructs)
Constructs:
{'dimensioncoordinate0': <DimensionCoordinate: latitude(5) degrees_north>,
 'dimensioncoordinate1': <DimensionCoordinate: longitude(8) degrees_east>,
 'dimensioncoordinate2': <DimensionCoordinate: time(1) days since 2018-12-01 >,
 'domainaxis0': <DomainAxis: size(5)>,
 'domainaxis1': <DomainAxis: size(8)>,
 'domainaxis2': <DomainAxis: size(1)>}

        '''
        return self._get_component('constructs')
    #--- End: def

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def copy(self, data=True):
        '''Return a deep copy.

``d.copy()`` is equivalent to ``copy.deepcopy(d)``.

.. versionadded:: 1.7.0

.. seealso:: `fromconstructs`

:Parameters:

    data: `bool`, optional
        If False then do not copy data. By default data are copied.

:Returns:	

        The deep copy.

**Examples:**

>>> e = d.copy()

        '''
        return type(self)(source=self, copy=True, _use_data=data)
    #--- End: def

    @classmethod
    def fromconstructs(cls, constructs, copy=False):
        '''Create a domain from existing metadata constructs.

The new domain act as a view to the given constructs, i.e. changes to
the domain, such as the addition or removal of a construct, will also
affect the input `Constructs` instance.

.. versionadded:: 1.7.0

:Parameters:

    constructs: `Constructs`
        The constructs from which to create the new domain. Cell
        method and field ancillary constucts are ignored.

    copy: bool, optional
        If True then deep copy the metadata constructs prior to
        initialization. By default the metadata constructs are not
        copied.

:Returns:

    `Domain`
        The domain created from a view of the constructs.

**Examples:**

>>> d = Domain.fromconstructs(f.constructs)

        '''
        domain = cls()
        domain._set_component('constructs',
                              constructs._view(ignore=('cell_method',
                                                       'field_ancillary')),
                              copy=copy)

        domain.constructs._field_data_axes = None
        
        return domain
    #--- End: def
            
#--- End: class



from builtins import object

import re


class ConstructAccess(object):
    '''Mixin class for manipulating constructs stored in a `Constructs`
object.

.. versionadded:: 1.7.0

    '''
    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    def _unique_construct_names(self):
        '''Return unique metadata construct names.

.. versionadded:: 1.7.0

**Examples:**

>>> f._unique_construct_names()
{'cellmethod0': 'method:mean',
 'dimensioncoordinate0': 'latitude',
 'dimensioncoordinate1': 'longitude',
 'dimensioncoordinate2': 'time',
 'domainaxis0': 'ncdim%lat',
 'domainaxis1': 'ncdim%lon',
 'domainaxis2': 'key%domainaxis2'}

        '''    
        key_to_name = {}
        name_to_keys = {}


        for d in self.constructs._constructs.values():
            name_to_keys = {}
        
            for key, construct in d.items():
                name = construct.identity(default='key%'+key)
                name_to_keys.setdefault(name, []).append(key)
                key_to_name[key] = name
    
            for name, keys in name_to_keys.items():
                if len(keys) <= 1:
                    continue
                
                for key in keys:
                    key_to_name[key] = '{0}{{{1}}}'.format(
                        name,
                        re.findall('\d+$', key)[0])
        #--- End: for
        
        return key_to_name
    #--- End: def
    
    def _unique_domain_axis_identities(self):
        '''Return unique domain axis construct names.

.. versionadded:: 1.7.0

**Examples:**

>>> f._unique_domain_axis_identities()
{'domainaxis0': 'latitude(5)',
 'domainaxis1': 'longitude(8)',
 'domainaxis2': 'time(1)'}

        '''
        key_to_name = {}
        name_to_keys = {}

        for key, value in self.domain_axes.items():
            name_size = (self.constructs.domain_axis_identity(key),
                         value.get_size(''))
            name_to_keys.setdefault(name_size, []).append(key)
            key_to_name[key] = name_size

        for (name, size), keys in name_to_keys.items():
            if len(keys) == 1:
                key_to_name[keys[0]] = '{0}({1})'.format(name, size)
            else:
                for key in keys:                    
                    key_to_name[key] = '{0}{{{1}}}({2})'.format(name,
                                                                re.findall('\d+$', key)[0],
                                                                size)
        #--- End: for
        
        return key_to_name
    #--- End: def

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def coordinate_references(self):
        '''Return coordinate reference constructs.

.. versionadded:: 1.7.0

.. seealso:: `constructs`

:Returns:

    `Constructs`
        The constructs and their construct keys.


**Examples:**

>>> f.coordinate_references
Constructs:
{}

>>> f.coordinate_references
Constructs:
{'coordinatereference0': <CoordinateReference: atmosphere_hybrid_height_coordinate>,
 'coordinatereference1': <CoordinateReference: rotated_latitude_longitude>}

        '''
        return self.constructs.filter_by_type('coordinate_reference')
    #--- End: def

    @property
    def domain_axes(self):
        '''Return domain axis constructs.

.. versionadded:: 1.7.0

.. seealso:: `constructs`

:Returns:

    `Constructs`
        The domain axis constructs and their construct keys.

**Examples:**

>>> f.domain_axes
Constructs:
{}

>>> f.domain_axes
Constructs:
{'domainaxis0': <DomainAxis: size(1)>,
 'domainaxis1': <DomainAxis: size(10)>,
 'domainaxis2': <DomainAxis: size(9)>,
 'domainaxis3': <DomainAxis: size(1)>}

        '''
        return self.constructs.filter_by_type('domain_axis')
    #--- End: def

    @property
    def auxiliary_coordinates(self):
        '''Return auxiliary coordinate constructs.

.. versionadded:: 1.7.0

.. seealso:: `constructs`

:Returns:

    `Constructs`
        The auxiliary coordinate constructs and their construct keys.

**Examples:**

>>> f.auxiliary_coordinates
Constructs:
{}

>>> f.auxiliary_coordinates
Constructs:
{'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degrees_N>,
 'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degrees_E>,
 'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name:Grid latitude name(10) >}

        '''
        return self.constructs.filter_by_type('auxiliary_coordinate')
    #--- End: def

    @property
    def dimension_coordinates(self):
        '''Return dimension coordinate constructs.

.. versionadded:: 1.7.0

.. seealso:: `constructs`

:Returns:

    `Constructs`
        The dimension coordinate constructs and their construct keys.

**Examples:**

>>> f.dimension_coordinates
Constructs:
{}

>>> f.dimension_coordinates
Constructs:
{'dimensioncoordinate0': <DimensionCoordinate: atmosphere_hybrid_height_coordinate(1) >,
 'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>,
 'dimensioncoordinate2': <DimensionCoordinate: grid_longitude(9) degrees>,
 'dimensioncoordinate3': <DimensionCoordinate: time(1) days since 2018-12-01 >}

        '''
        return self.constructs.filter_by_type('dimension_coordinate')
    #--- End: def
    
    @property
    def coordinates(self):
        '''Return dimension and auxiliary coordinate constructs.

.. versionadded:: 1.7.0

.. seealso:: `auxiliary_coordinates`, `constructs`,
             `dimension_coordinates`

:Returns:

    `Constructs`
        The auxiliary coordinate and dimension coordinate constructs
        and their construct keys.

**Examples:**

>>> f.coordinates
Constructs:
{}

>>> f.coordinates
Constructs:
{'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degrees_N>,
 'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degrees_E>,
 'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name=Grid latitude name(10) >,
 'dimensioncoordinate0': <DimensionCoordinate: atmosphere_hybrid_height_coordinate(1) >,
 'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>,
 'dimensioncoordinate2': <DimensionCoordinate: grid_longitude(9) degrees>,
 'dimensioncoordinate3': <DimensionCoordinate: time(1) days since 2018-12-01 >}

        '''
        out = self.dimension_coordinates
        out._update(self.auxiliary_coordinates)        
        return out
    #--- End: def

    @property
    def domain_ancillaries(self):
        '''Return domain ancillary constructs.

.. versionadded:: 1.7.0

.. seealso:: `constructs`

:Returns:

    `Constructs`
        The domain ancillary constructs and their construct keys.

**Examples:**

>>> f.domain_ancillaries
Constructs:
{}

>>> f.domain_ancillaries
Constructs:
{'domainancillary0': <DomainAncillary: ncvar%a(1) m>,
 'domainancillary1': <DomainAncillary: ncvar%b(1) >,
 'domainancillary2': <DomainAncillary: surface_altitude(10, 9) m>}

        '''
        return self.constructs.filter_by_type('domain_ancillary')
    #--- End: def

    @property
    def cell_measures(self):
        '''Return cell measure constructs.

.. versionadded:: 1.7.0

.. seealso:: `constructs`

:Returns:

    `Constructs`
        The cell measure constructs and their construct keys.

**Examples:**

>>> f.cell_measures
Constructs:
{}

>>> f.cell_measures
Constructs:
{'cellmeasure0': <CellMeasure: measure%area(9, 10) km2>}

        '''
        return self.constructs.filter_by_type('cell_measure')
    #--- End: def

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def construct(self, identity, default=ValueError()):
        '''Select a metadata construct by its identity.

.. versionadded:: 1.7.0

.. seealso:: `construct_key`, `constructs`,
             `Constructs.filter_by_identity`, `Constructs.value`

:Parameters:

    identity: optional

        Select constructs that have the given identity. If exactly one
        construct is selected then it is returned, otherwise an
        exception is raised.

        The identity is specified by a string (e.g. ``'latitude'``,
        ``'long_name=time'``, etc.); or a compiled regular expression
        (e.g. ``re.compile('^atmosphere')``), for which all constructs
        whose identities match (via `re.search`) are selected.

        Each construct has a number of identities, and is selected if
        any of them match any of those provided. A construct's
        identities are those returned by its `!identities` method. In
        the following example, the construct ``c`` has four
        identities:

           >>> c.identities()
           ['time', 'long_name=Time', 'foo=bar', 'ncvar%T']

        In addition, each construct also has an identity based its
        construct key (e.g. ``'key%dimensioncoordinate2'``)

        Note that in the output of a `print` call or `!dump` method, a
        construct is always described by one of its identities, and so
        this description may always be used as an *identity* argument.

    default: optional
        Return the value of the *default* parameter if the property
        has not been set. If set to an `Exception` instance then it
        will be raised instead.

:Returns:

        The selected construct.

**Examples:**

>>> print(f.constructs)
Constructs:
{'cellmethod0': <CellMethod: area: mean>,
 'dimensioncoordinate0': <DimensionCoordinate: latitude(5) degrees_north>,
 'dimensioncoordinate1': <DimensionCoordinate: longitude(8) degrees_east>,
 'dimensioncoordinate2': <DimensionCoordinate: long_name=time(1) days since 2018-12-01 >,
 'domainaxis0': <DomainAxis: size(5)>,
 'domainaxis1': <DomainAxis: size(8)>,
 'domainaxis2': <DomainAxis: size(1)>}

Select the construct that has the "standard_name" property of 'latitude':

>>> f.construct('latitude')
<DimensionCoordinate: latitude(5) degrees_north>

Select the cell method construct that has a "method" of 'mean':

>>> f.construct('method:mean')
<CellMethod: area: mean>

Attempt to select the construct whose "standard_name" start with the
letter 'l':

>>> import re
>>> f.construct(re.compile('^l'))
ValueError: Can't return 2 constructs
>>> f.construct(re.compile('^l'), default='no construct')
'no construct'

        '''
        c = self.constructs.filter_by_identity(identity)
        if len(c) == 1:
            return c.value()

        if not c:
            return self._default(default,
                                 "Can't return zero constructs")

        return self._default(default,
                             "Can't return {0} constructs".format(len(c)))
    #--- End: def

    def construct_key(self, identity, default=ValueError()):
        '''Select the key of a metadata construct by its identity.

.. versionadded:: 1.7.0

.. seealso:: `construct`, `constructs`,
             `Constructs.filter_by_identity`, `Constructs.key`

:Parameters:

    identity: optional

        Select constructs that have the given identity. If exactly one
        construct is selected then it is returned, otherwise an
        exception is raised.

        The identity is specified by a string (e.g. ``'latitude'``,
        ``'long_name=time'``, etc.); or a compiled regular expression
        (e.g. ``re.compile('^atmosphere')``), for which all constructs
        whose identities match (via `re.search`) are selected.

        Each construct has a number of identities, and is selected if
        any of them match any of those provided. A construct's
        identities are those returned by its `!identities` method. In
        the following example, the construct ``c`` has four
        identities:

           >>> c.identities()
           ['time', 'long_name=Time', 'foo=bar', 'ncvar%T']

        In addition, each construct also has an identity based its
        construct key (e.g. ``'key%dimensioncoordinate2'``)

        Note that in the output of a `print` call or `!dump` method, a
        construct is always described by one of its identities, and so
        this description may always be used as an *identity* argument.

    default: optional
        Return the value of the *default* parameter if the property
        has not been set. If set to an `Exception` instance then it
        will be raised instead.

:Returns:

    `str`
        The key of the selected construct.

**Examples:**

>>> print(f.constructs)
Constructs:
{'cellmethod0': <CellMethod: area: mean>,
 'dimensioncoordinate0': <DimensionCoordinate: latitude(5) degrees_north>,
 'dimensioncoordinate1': <DimensionCoordinate: longitude(8) degrees_east>,
 'dimensioncoordinate2': <DimensionCoordinate: long_name=time(1) days since 2018-12-01 >,
 'domainaxis0': <DomainAxis: size(5)>,
 'domainaxis1': <DomainAxis: size(8)>,
 'domainaxis2': <DomainAxis: size(1)>}

Select the construct that has the "standard_name" property of 'latitude':

>>> f.construct_key('latitude')
 'dimensioncoordinate0'

Select the cell method construct that has a "method" of 'mean':

>>> f.construct_key('method:mean')
'cellmethod0'

Attempt to select the construct whose "standard_name" start with the
letter 'l':

>>> import re
>>> f.construct_key(re.compile('^l'))
ValueError: Can't return the key of 2 constructs
>>> f.construct_key(re.compile('^l'), default='no construct')
'no construct'

        '''
        c = self.constructs.filter_by_identity(identity)
        if len(c) == 1:
            return c.key()

        if not c:
            return self._default(default,
                                 "Can't return the key of zero constructs")

        return self._default(
            default,
            "Can't return the keys of {0} constructs".format(len(c)))
    #--- End: def

    def domain_axis_key(self, identity, default=ValueError()):
        '''Return the key of the domain axis construct that is spanned by 1-d
coordinate constructs.

:Parameters:

    identity:

        Select the 1-d coordinate constructs that have the given
        identity.

        An identity is specified by a string (e.g. ``'latitude'``,
        ``'long_name=time'``, etc.); or a compiled regular expression
        (e.g. ``re.compile('^atmosphere')``), for which all constructs
        whose identities match (via `re.search`) are selected.

        Each coordinate construct has a number of identities, and is
        selected if any of them match any of those provided. A
        construct's identities are those returned by its `!identities`
        method. In the following example, the construct ``x`` has four
        identities:

           >>> x.identities()
           ['time', 'long_name=Time', 'foo=bar', 'ncvar%T']

        In addition, each construct also has an identity based its
        construct key (e.g. ``'key%dimensioncoordinate2'``)

        Note that in the output of a `print` call or `!dump` method, a
        construct is always described by one of its identities, and so
        this description may always be used as an *identity* argument.

    default: optional
        Return the value of the *default* parameter if a domain axis
        construct can not be found. If set to an `Exception` instance
        then it will be raised instead.

:Returns:

    `str`
        The key of the domain axis construct that is spanned by the
        data of the selected 1-d coordinate constructs.

**Examples:**

TODO

        '''
        constructs = self.constructs

        # Select 1-d coordinate constructs with the given identity
        c = constructs.filter_by_type('dimension_coordinate',
                                      'auxiliary_coordinates')
        c = c.filter_by_naxes(1)
        c = c.filter_by_identity(identity)
        
        if not len(c) :
            return self._default(
                default,
                "No 1-d coordinate constructs have identity {!r}".format(identity))

        data_axes = constructs.data_axes()
        domain_axes = constructs.filter_by_type('domain_axis')
        
        keys = []
        for ckey, coord in c.items():
            axes = data_axes.get(ckey)
            if not axes:
                continue
            
            key = axes[0]
            if domain_axes.get(key):
                keys.append(key)
        #--- End: for
        
        keys = set(keys)
            
        if not keys:
            return self._default(
                default,
                "1-d coordinate constructs selected with identity {!r} have not been assigned a domain axis contructs".format(
                    coord))                       
        
        if len(keys) > 1:
            return self._default(
                default,
                "Multiple 1-d coordinate constructs selected with identity {!r} span multiple domain axes: {!r}".format(
                    identity, keys))

        return keys.pop()
    #--- End: def

        
#--- End: class



from __future__ import print_function

import datetime
import inspect
import os
import unittest

import numpy

import cfdm

class CoordinateReferenceTest(unittest.TestCase):
    def setUp(self):
        self.filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                     'test_file.nc')
        f = cfdm.read(self.filename)
        self.assertTrue(len(f)==1, 'f={!r}'.format(f))
        self.f = f[0]

        self.test_only = []
    #--- End: def

    def test_CoordinateReference__repr__str__dump_construct_type(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f

        for cr in f.coordinate_references.values():
            _ = repr(cr)
            _ = str(cr)
            _ = cr.dump(display=False)
            self.assertTrue(cr.construct_type == 'coordinate_reference')
    #--- End: def

    def test_CoordinateReference_equals(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f.copy()

        # Create a vertical grid mapping coordinate reference
        t = cfdm.CoordinateReference(
            coordinates=('coord1',),
            coordinate_conversion=cfdm.CoordinateConversion(
                parameters={'standard_name': 'atmosphere_hybrid_height_coordinate'},
                domain_ancillaries={'a': 'aux0', 'b': 'aux1', 'orog': 'orog'})
        )
        self.assertTrue(t.equals(t.copy(), verbose=True))
        
        # Create a horizontal grid mapping coordinate reference
        t = cfdm.CoordinateReference(
            coordinates=['coord1', 'fred', 'coord3'],
             coordinate_conversion=cfdm.CoordinateConversion(
                 parameters={'grid_mapping_name': 'rotated_latitude_longitude',
                             'grid_north_pole_latitude': 38.0,
                             'grid_north_pole_longitude': 190.0})
        )            
        self.assertTrue(t.equals(t.copy(), verbose=True))

        datum=cfdm.Datum(parameters={'earth_radius': 6371007})
        conversion=cfdm.CoordinateConversion(
            parameters={'grid_mapping_name': 'rotated_latitude_longitude',
                        'grid_north_pole_latitude': 38.0,
                        'grid_north_pole_longitude': 190.0})
        
        t = cfdm.CoordinateReference(
            coordinate_conversion=conversion,
            datum=datum,
            coordinates=['x', 'y', 'lat', 'lon']
        )

        self.assertTrue(t.equals(t.copy(), verbose=True))

        # Create a horizontal grid mapping coordinate reference
        t = cfdm.CoordinateReference(
               coordinates=['coord1', 'fred', 'coord3'],
               coordinate_conversion=cfdm.CoordinateConversion(
                   parameters={'grid_mapping_name': 'albers_conical_equal_area',
                               'standard_parallel': [-30, 10],
                               'longitude_of_projection_origin': 34.8,
                               'false_easting': -20000,
                               'false_northing': -30000})
        )
        self.assertTrue(t.equals(t.copy(), verbose=True))

        # Create a horizontal grid mapping coordinate reference
        t = cfdm.CoordinateReference(
               coordinates=['coord1', 'fred', 'coord3'],
               coordinate_conversion=cfdm.CoordinateConversion(
                   parameters={'grid_mapping_name': 'albers_conical_equal_area',
                               'standard_parallel': cfdm.Data([-30, 10]),
                               'longitude_of_projection_origin': 34.8,
                               'false_easting': -20000,
                               'false_northing': -30000})
        )
        self.assertTrue(t.equals(t.copy(), verbose=True))
    #--- End: def

    def test_CoordinateConversion(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f.copy()

        cr = f.construct('standard_name:atmosphere_hybrid_height_coordinate')
        cc = cr.coordinate_conversion
        _ = repr(cc)
        _ = str(cc)

        domain_ancillaries = cc.domain_ancillaries()
        self.assertTrue(len(domain_ancillaries) == 3)
        
        for key, value in domain_ancillaries.items():
            self.assertTrue(cc.has_domain_ancillary(key))
            self.assertTrue(cc.get_domain_ancillary(key) == value)
            _ = cc.del_domain_ancillary(key)
            self.assertFalse(cc.has_domain_ancillary(key))
            self.assertTrue(cc.get_domain_ancillary(key, None) == None)
            self.assertTrue(cc.del_domain_ancillary(key, None) == None)
            cc.set_domain_ancillary(key, _)
            self.assertTrue(cc.has_domain_ancillary(key))
            self.assertTrue(cc.get_domain_ancillary(key) == value)

        cr = f.construct('grid_mapping_name:rotated_latitude_longitude')
        cc = cr.coordinate_conversion
        _ = repr(cc)
        _ = str(cc)

        parameters = cc.parameters()
        self.assertTrue(len(parameters) == 3, parameters)
        
        for key, value in parameters.items():
            self.assertTrue(cc.has_parameter(key))
            self.assertTrue(cc.get_parameter(key) == value)
            _ = cc.del_parameter(key)
            self.assertFalse(cc.has_parameter(key))
            self.assertTrue(cc.get_parameter(key, None) == None)
            self.assertTrue(cc.del_parameter(key, None) == None)
            cc.set_parameter(key, _)
            self.assertTrue(cc.has_parameter(key))
            self.assertTrue(cc.get_parameter(key) == value)

        _ = cr.del_coordinate_conversion()
        self.assertTrue(_.equals(cc))
    #--- End: def

    def test_Datum(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f.copy()

        cr = f.construct('standard_name:atmosphere_hybrid_height_coordinate')
        d = cr.datum
        _ = repr(d)
        _ = str(d)

        parameters = d.parameters()
        self.assertTrue(len(parameters) == 1, parameters)
        
        for key, value in parameters.items():
            self.assertTrue(d.has_parameter(key))
            self.assertTrue(d.get_parameter(key) == value)
            _ = d.del_parameter(key)
            self.assertFalse(d.has_parameter(key))
            self.assertTrue(d.get_parameter(key, None) == None)
            self.assertTrue(d.del_parameter(key, None) == None)
            d.set_parameter(key, _)
            self.assertTrue(d.has_parameter(key))
            self.assertTrue(d.get_parameter(key) == value)

        _ = cr.del_datum()
        self.assertTrue(_.equals(d))

        f = self.f.copy()
        key = f.construct_key('grid_mapping_name:rotated_latitude_longitude')
        f.del_construct(key)
        cr = f.construct('standard_name:atmosphere_hybrid_height_coordinate')
        cr.datum.nc_set_variable('my_name')
        cfdm.write(f, 'delme.nc')
    #--- End: def

#--- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    print(cfdm.environment(display=False))
    print('')
    unittest.main(verbosity=2)



from __future__ import print_function

import datetime
import inspect
import os
import unittest

import numpy

import cfdm

class CellMethodTest(unittest.TestCase):
    def setUp(self):
        self.filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                     'test_file.nc')
        f = cfdm.read(self.filename)
        self.assertTrue(len(f)==1, 'f={!r}'.format(f))
        self.f = f[0]

        self.test_only = []
    #--- End: def

    def test_CellMethod__repr__str__dump_construct_type(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f

        for c in f.cell_methods.values():
            _ = repr(c)
            _ = str(c)
            _ = c.dump(display=False)
            self.assertTrue(c.construct_type == 'cell_method')
    #--- End: def

    def test_CellMethod(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f

        # ------------------------------------------------------------
        # Equals and idenities
        # ------------------------------------------------------------
        for c in f.cell_methods.values():
            d = c.copy()
            self.assertTrue(c.equals(c, verbose=True))
            self.assertTrue(c.equals(d, verbose=True))
            self.assertTrue(d.equals(c, verbose=True))
            self.assertTrue(c.identity() == 'method:'+c.get_method())
            self.assertTrue(c.identities() == ['method:'+c.get_method()])

        # ------------------------------------------------------------
        # Sorted
        # ------------------------------------------------------------
        c = cfdm.CellMethod(method='minimum',
                            axes=['B', 'A'],
                            qualifiers={'interval': [1, 2]})
        
        d = cfdm.CellMethod(method='minimum',
                            axes=['A', 'B'],
                            qualifiers={'interval': [2, 1]})
        
        self.assertTrue(d.equals(c.sorted(), verbose=True))

        c = cfdm.CellMethod(method='minimum',
                            axes=['B', 'A'],
                            qualifiers={'interval': [3]})
        
        d = cfdm.CellMethod(method='minimum',
                            axes=['A', 'B'],
                            qualifiers={'interval': [3]})
        
        self.assertTrue(d.equals(c.sorted(), verbose=True))

        c = cfdm.CellMethod(method='minimum',
                            axes=['area'],
                            qualifiers={'interval': [3]})
        
        d = cfdm.CellMethod(method='minimum',
                            axes=['area'],
                            qualifiers={'interval': [3]})
        
        self.assertTrue(d.equals(c.sorted(), verbose=True))
   #--- End: def
#--- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    print(cfdm.environment(display=False))
    print('')
    unittest.main(verbosity=2)



from future.utils import with_metaclass
from builtins import (str, super)

import abc

from copy import deepcopy

from . import Container


class Properties(with_metaclass(abc.ABCMeta, Container)):
    '''Abstract base class for an object with descriptive properties.

.. versionadded:: 1.7.0

    '''
    def __init__(self, properties=None, source=None, copy=True):
        '''**Initialization**

:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
           ``properties={'standard_name': 'altitude'}``
        
        Properties may also be set after initialisation with the
        `set_properties` and `set_property` methods.

    source: optional
        Initialize the properties from those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization By default parameters are deep copied.

        '''
        super().__init__(source=source, copy=copy)
        
        if source is not None:
            try:
                properties = source.properties()
            except AttributeError:
                properties = None
        #--- End: if

        self._set_component('properties', {}, copy=False)
        
        if properties is not None:
            self.set_properties(properties, copy=copy)
    #--- End: def

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def clear_properties(self):
        '''Remove all properties.

    .. versionadded:: 1.7.0
    
    .. seealso:: `del_property` `properties`, `set_properties`
    
    :Returns:
    
        `dict`
            The properties that have been removed.
    
    **Examples:**

    >>> f.properties()
    {}
    >>> f.set_properties({'standard_name': 'air_pressure', 'long_name': 'Air Pressure'})
    >>> f.properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.set_properties({'standard_name': 'air_pressure', 'foo': 'bar'})
    >>> f.properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.clear_properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.properties()
    {}

        '''
        out = self._get_component('properties')
        self._set_component('properties', {})
        return out.copy()

    
    def del_property(self, prop, default=ValueError()):
        '''Remove a property.

.. versionadded:: 1.7.0

.. seealso:: `clear_properties`, `get_property`, `has_property`,
             `properties`, `set_property`

:Parameters:

    prop: `str`
        The name of the property to be removed.

        *Parameter example:*
           ``prop='long_name'``

    default: optional
        Return the value of the *default* parameter if the property
        has not been set. If set to an `Exception` instance then it
        will be raised instead.

:Returns:

        The removed property.

**Examples:**

>>> f.set_property('project', 'CMIP7')
>>> f.has_property('project')
True
>>> f.get_property('project')
'CMIP7'
>>> f.del_property('project')
'CMIP7'
>>> f.has_property('project')
False
>>> print(f.del_property('project', None))
None
>>> print(f.get_property('project', None))
None

        '''
        try:
            return self._get_component('properties').pop(prop)
        except KeyError:
            return self._default(default,
                                 "{!r} has no {!r} property".format(
                                 self.__class__.__name__, prop))
    #--- End: def

    def get_property(self, prop, default=ValueError()):
        '''Return a property.

.. versionadded:: 1.7.0

.. seealso:: `del_property`, `has_property`, `properties`,
             `set_property`

:Parameters:

    prop: `str`
        The name of the property to be returned.

        *Parameter example:*
           ``prop='standard_name'``

    default: optional
        Return the value of the *default* parameter if the property
        has not been set. If set to an `Exception` instance then it
        will be raised instead.

:Returns:

        The value of the property.

**Examples:**

>>> f.set_property('project', 'CMIP7')
>>> f.has_property('project')
True
>>> f.get_property('project')
'CMIP7'
>>> f.del_property('project')
'CMIP7'
>>> f.has_property('project')
False
>>> print(f.del_property('project', None))
None
>>> print(f.get_property('project', None))
None

        '''
        try:
            return self._get_component('properties')[prop]
        except KeyError:
            return self._default(default,
                                 "{!r} has no {!r} property".format(
                                     self.__class__.__name__, prop))
    #--- End: def

    def has_property(self, prop):
        '''Whether a property has been set.

.. versionadded:: 1.7.0

.. seealso:: `del_property`, `get_property`, `properties`,
             `set_property`

:Parameters:

    prop: `str`
        The name of the property.

        *Parameter example:*
           ``prop='long_name'``

:Returns:

     `bool`
        True if the property has been set, otherwise False.

**Examples:**

>>> f.set_property('project', 'CMIP7')
>>> f.has_property('project')
True
>>> f.get_property('project')
'CMIP7'
>>> f.del_property('project')
'CMIP7'
>>> f.has_property('project')
False
>>> print(f.del_property('project', None))
None
>>> print(f.get_property('project', None))
None

        '''
        return prop in self._get_component('properties')
    #--- End: def

    def properties(self):
        '''Return all properties.

    .. versionadded:: 1.7.0
    
    .. seealso:: `clear_properties`, `get_property`, `has_property`
                 `set_properties`
    
    :Returns:
    
        `dict`
            The properties.
    
    **Examples:**
    
    >>> f.properties()
    {}
    >>> f.set_properties({'standard_name': 'air_pressure', 'long_name': 'Air Pressure'})
    >>> f.properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.set_properties({'standard_name': 'air_pressure', 'foo': 'bar'})
    >>> f.properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.clear_properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.properties()
    {}

        '''
        return self._get_component('properties').copy()


    def set_properties(self, properties, copy=True):
        '''Set properties.

    .. versionadded:: 1.7.0
    
    .. seealso:: `clear_properties`, `properties`, `set_property`
    
    :Parameters:
    
        properties: `dict` 
            Store the properties from the dictionary supplied.
    
            *Parameter example:*

              ``properties={'standard_name': 'altitude', 'foo': 'bar'}``
    
        copy: `bool`, optional
            If False then any property values provided by the
            *properties* parameter are not copied before insertion. By
            default they are deep copied.
    
    :Returns:
    
        `None`
    
    **Examples:**
    
    >>> f.properties()
    {}
    >>> f.set_properties({'standard_name': 'air_pressure', 'long_name': 'Air Pressure'})
    >>> f.properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.set_properties({'standard_name': 'air_pressure', 'foo': 'bar'})
    >>> f.properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.clear_properties()
    {'standard_name': 'air_pressure',
     'foo': 'bar',
     'long_name': 'Air Pressure'}
    >>> f.properties()
    {}

        '''
        if copy:
            properties = deepcopy(properties)                
        else:
            properties = properties.copy()
        
        self._get_component('properties').update(properties)


    def set_property(self, prop, value, copy=True):
        '''Set a property.

.. versionadded:: 1.7.0

.. seealso:: `del_property`, `get_property`, `has_property`,
             `properties`, `set_properties`

:Parameters:

    prop: `str`
        The name of the property to be set.

    value:
        The value for the property.

    copy: `bool`, optional
        If True then set a deep copy of *value*.

:Returns:

     `None`

**Examples:**

>>> f.set_property('project', 'CMIP7')
>>> f.has_property('project')
True
>>> f.get_property('project')
'CMIP7'
>>> f.del_property('project')
'CMIP7'
>>> f.has_property('project')
False
>>> print(f.del_property('project', None))
None
>>> print(f.get_property('project', None))
None

        '''
        if copy:
            value = deepcopy(value)
            
        self._get_component('properties')[prop] = value
    #--- End: def

#--- End: class



import sys

import numpy

CONSTANTS = {
    # The value of absolute tolerance for testing numerically tolerant
    # equality.
    'RTOL': sys.float_info.epsilon,
    # The value of relative tolerance for testing numerically tolerant
    # equality.
    'ATOL': sys.float_info.epsilon,
}

# --------------------------------------------------------------------
# masked
# --------------------------------------------------------------------
'''A constant that allows data values to be masked by direct
assignment. This is consistent with the behaviour of numpy masked
arrays.

For example, masking every element of a field constructs data array
could be done as follows:

>>> f[...] = cfdm.masked

'''
masked = numpy.ma.masked



from __future__ import print_function
from builtins import super

import textwrap

import numpy

from . import Container


class Properties(Container):
    '''Mixin class for descriptive properties.

    .. versionadded:: 1.7.0

    '''
    def __str__(self):
        '''Called by the `str` built-in function.

    x.__str__() <==> str(x)

    .. versionadded:: 1.7.0

        '''
        return '{0}'.format(self.identity(''))


    def _dump_properties(self, _prefix='', _level=0,
                         _omit_properties=None):
        '''Dump the properties.

    .. versionadded:: 1.7.0
    
    :Parameters:
    
        _omit_properties: sequence of `str`, optional
            Omit the given CF properties from the description.
    
        _level: `int`, optional
    
    :Returns:
    
        `str`

        '''
        indent0 = '    ' * _level
        string = []

        properties = self.properties()
        
        if _omit_properties:
            for prop in _omit_properties:
                 properties.pop(prop, None)
        #--- End: if
 
        for prop, value in sorted(properties.items()):
            name   = '{0}{1}{2} = '.format(indent0, _prefix, prop)
            value  = repr(value)
            subsequent_indent = ' ' * len(name)
            if value.startswith("'") or value.startswith('"'):
                subsequent_indent = '{0} '.format(subsequent_indent)
                
            string.append(
                textwrap.fill(name+value, 79,
                              subsequent_indent=subsequent_indent))
        
        return '\n'.join(string)


    def dump(self, display=True, _key=None, _omit_properties=(),
             _prefix='', _title=None, _create_title=True, _level=0):
        '''A full description.

    .. versionadded:: 1.7.0
    
    :Parameters:
    
        display: `bool`, optional
            If False then return the description as a string. By
            default the description is printed.
    
    :Returns:
    
            The description. If *display* is True then the description
            is printed and `None` is returned. Otherwise the
            description is returned as a string.

        '''
        indent0 = '    ' * _level
        indent1 = '    ' * (_level+1)

        string = []

        # ------------------------------------------------------------
        # Title
        # ------------------------------------------------------------
        if _create_title:
            if _title is None:
                if _key:
                    default = 'key%{0}'.format(_key)
                else:
                    default = ''
                    
                string.append('{0}{1}: {2}'.format(indent0,
                                                   self.__class__.__name__,
                                                   self.identity(default=default)))
            else:
                string.append(indent0 + _title)
        #--- End: if
        
        # ------------------------------------------------------------
        # Properties
        # ------------------------------------------------------------
        properties = self._dump_properties(_prefix=_prefix,
                                           _level=_level+1,
                                           _omit_properties=_omit_properties)
        if properties:
            string.append(properties)

        string = '\n'.join(string)
       
        if display:
            print(string)
        else:
            return string


    def equals(self, other, rtol=None, atol=None, verbose=False,
               ignore_data_type=False, ignore_fill_value=False,
               ignore_properties=(), ignore_type=False):
        '''Whether two instances are the same.

    Equality is strict by default. This means that:
    
    * the same descriptive properties must be present, with the same
      values and data types, and vector-valued properties must also have
      same the size and be element-wise equal (see the *ignore_properties*
      and *ignore_data_type* parameters).
    
    Two real numbers ``x`` and ``y`` are considered equal if
    ``|x-y|<=atol+rtol|y|``, where ``atol`` (the tolerance on absolute
    differences) and ``rtol`` (the tolerance on relative differences) are
    positive, typically very small numbers. See the *atol* and *rtol*
    parameters.
    
    Any type of object may be tested but, in general, equality is only
    possible with another object of the same type, or a subclass of
    one. See the *ignore_type* parameter.
    
    .. versionadded:: 1.7.0
    
    :Parameters:
    
        other: 
            The object to compare for equality.
    
        atol: float, optional
            The tolerance on absolute differences between real
            numbers. The default value is set by the `cfdm.ATOL`
            function.
            
        rtol: float, optional
            The tolerance on relative differences between real
            numbers. The default value is set by the `cfdm.RTOL`
            function.
    
        ignore_fill_value: `bool`, optional
            If True then the "_FillValue" and "missing_value"
            properties are omitted from the comparison.
    
        verbose: `bool`, optional
            If True then print information about differences that lead
            to inequality.
    
        ignore_properties: sequence of `str`, optional
            The names of properties to omit from the comparison.
    
        ignore_data_type: `bool`, optional
            If True then ignore the data types in all numerical
            comparisons. By default different numerical data types
            imply inequality, regardless of whether the elements are
            within the tolerance for equality.
    
        ignore_type: `bool`, optional
            Any type of object may be tested but, in general, equality
            is only possible with another object of the same type, or
            a subclass of one. If *ignore_type* is True then equality
            is possible for any object with a compatible API.
    
    :Returns: 
      
        `bool`
            Whether the two instances are equal.
    
    **Examples:**
    
    >>> p.equals(p)
    True
    >>> p.equals(p.copy())
    True
    >>> p.equals('not a colection of properties')
    False
    
    >>> q = p.copy()
    >>> q.set_property('foo', 'bar')
    >>> p.equals(q)
    False
    >>> p.equals(q, verbose=True)
    Field: Non-common property name: foo
    Field: Different properties
    False

        '''
        pp = super()._equals_preprocess(other, verbose=verbose,
                                        ignore_type=ignore_type)
        if pp is True or pp is False:
            return pp
        
        other = pp
        
        # ------------------------------------------------------------
        # Check the properties
        # ------------------------------------------------------------
        if ignore_fill_value:
            ignore_properties += ('_FillValue', 'missing_value')

        self_properties  = self.properties()
        other_properties = other.properties()

        if ignore_properties:
            for prop in ignore_properties:
                self_properties.pop(prop, None)
                other_properties.pop(prop, None)
        #--- End: if
                
        if set(self_properties) != set(other_properties):
            if verbose:
                for prop in set(self_properties).symmetric_difference(other_properties):
                    print("{0}: Missing property: {1}".format( 
                        self.__class__.__name__, prop))
            return False

        for prop, x in self_properties.items():
            y = other_properties[prop]

            if not self._equals(x, y,
                                rtol=rtol, atol=atol,
                                ignore_fill_value=ignore_fill_value,
                                ignore_data_type=True, #ignore_data_type,
                                verbose=verbose):
                if verbose:
                    print("{0}: Different {1!r} property values: {2!r}, {3!r}".format(
                        self.__class__.__name__, prop, x, y))
                return False
        #--- End: for

        return True


    def identity(self, default=''):
        '''Return the canonical identity.

    By default the identity is the first found of the following:
    
    * The "standard_name" property.
    * The "cf_role" property, preceeded by ``'cf_role='``.
    * The "axis" property, preceeded by ``'axis='``.
    * The "long_name" property, preceeded by ``'long_name='``.
    * The netCDF variable name, preceeded by ``'ncvar%'``.
    * The value of the *default* parameter.
    
    .. versionadded:: 1.7.0
    
    .. seealso:: `identities`
    
    :Parameters:
    
        default: optional
            If no identity can be found then return the value of the
            default parameter.
    
    :Returns:
    
            The identity.
    
    **Examples:**
    
    >>> f.properties()
    {'foo': 'bar',
     'long_name': 'Air Temperature',
     'standard_name': 'air_temperature'}
    >>> f.nc_get_variable()
    'tas'
    >>> f.identity()
    'air_temperature'
    >>> f.del_property('standard_name')
    'air_temperature'
    >>> f.identity(default='no identity')
    'air_temperature'
    >>> f.identity()
    'long_name=Air Temperature'
    >>> f.del_property('long_name')
    >>> f.identity()
    'ncvar%tas'
    >>> f.nc_del_variable()
    'tas'
    >>> f.identity()
    'ncvar%tas'
    >>> f.identity()
    ''
    >>> f.identity(default='no identity')
    'no identity'

        '''
        n = self.get_property('standard_name', None)
        if n is not None:
            return n
        
        for prop in  ('cf_role', 'axis', 'long_name'):
            n = self.get_property(prop, None)
            if n is not None:
                return '{0}={1}'.format(prop, n)
        #--- End: for
        
        n = self.nc_get_variable(None)
        if n is not None:
            return 'ncvar%{0}'.format(n)
        
        return default


    def identities(self):
        '''Return all possible identities.

    The identities comprise:
    
    * The "standard_name" property.
    * All properties, preceeded by the property name and a colon,
      e.g. ``'long_name:Air temperature'``.
    * The netCDF variable name, preceeded by ``'ncvar%'``.
    
    .. versionadded:: 1.7.0
    
    .. seealso:: `identity`
    
    :Returns:
    
        `list`
            The identities.
    
    **Examples:**
    
    >>> f.properties()
    {'foo': 'bar',
     'long_name': 'Air Temperature',
     'standard_name': 'air_temperature'}
    >>> f.nc_get_variable()
    'tas'
    >>> f.identities()
    ['air_temperature',
     'long_name=Air Temperature',
     'foo=bar',
     'standard_name=air_temperature',
     'ncvar%tas']

        '''
        properties = self.properties()
        cf_role = properties.pop('cf_role', None)
        axis = properties.pop('axis', None)
        long_name = properties.pop('long_name', None)
        standard_name = properties.pop('standard_name', None)

        out = []

        if standard_name is not None:
            out.append(standard_name)
            
        if cf_role is not None:
            out.append('cf_role={}'.format(cf_role))
                   
        if axis is not None:
            out.append('axis={}'.format(axis))
                   
        if long_name is not None:
            out.append('long_name={}'.format(long_name))
            
        out += ['{0}={1}'.format(prop, value)
                for prop, value in sorted(properties.items())]

        if standard_name is not None:
            out.append('standard_name={}'.format(standard_name))
            
        n = self.nc_get_variable(None)
        if n is not None:
            out.append('ncvar%{0}'.format(n))
            
        return out


#--- End: class



from __future__ import print_function
import datetime
import inspect
import unittest

import cfdm

class FunctionsTest(unittest.TestCase):
    def setUp(self):
        self.test_only = []
    #--- End: def
        
    def test_CF_ATOL_RTOL_environment(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        self.assertTrue(cfdm.CF() == cfdm.core.__cf_version__)

        org = cfdm.RTOL()
        self.assertTrue(cfdm.RTOL(1e-5) == org)
        self.assertTrue(cfdm.RTOL() == 1e-5)
        self.assertTrue(cfdm.RTOL(org) == 1e-5)
        self.assertTrue(cfdm.RTOL() == org)
        
        org = cfdm.ATOL()
        self.assertTrue(cfdm.ATOL(1e-5) == org)
        self.assertTrue(cfdm.ATOL() == 1e-5)
        self.assertTrue(cfdm.ATOL(org) == 1e-5)
        self.assertTrue(cfdm.ATOL() == org)

        out = cfdm.environment(display=False)
    #--- End: def
    
#--- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    print(cfdm.environment(display=False))
    print('')
    unittest.main(verbosity=2)



from builtins import object
#from future.utils import with_metaclass

#import abc


#class ConstructAccess(with_metaclass(abc.ABCMeta, object)):
class ConstructAccess(object):
    '''Mixin class for accessing an embedded `Constructs` object.

.. versionadded:: 1.7.0

    '''   
    def del_construct(self, key, default=ValueError()):
        '''Remove a metadata construct.

If a domain axis construct is selected for removal then it can't be
spanned by any metdata construct data, nor the field construct's
data; nor be referenced by any cell method constructs.

However, a domain ancillary construct may be removed even if it is
referenced by coordinate reference construct. In this case the
reference is replace with `None`.

.. versionadded:: 1.7.0

.. seealso:: `constructs`, `get_construct`, `has_construct`,
             `set_construct`

:Parameters:

    key: `str`, optional
        The key of the metadata construct to be removed.

        *Parameter example:*
          ``key='auxiliarycoordinate0'``
        
    default: optional
        Return the value of the *default* parameter if the construct
        can not be removed, or does not exist. If set to an
        `Exception` instance then it will be raised instead.

:Returns:

        The removed metadata construct.

**Examples:**

>>> f.del_construct('dimensioncoordinate1')
<DimensionCoordinate: grid_latitude(111) degrees>

        '''
        return self.constructs._del_construct(key, default=default)
    #--- End: def

    def get_construct(self, key, default=ValueError()):
        '''Return a metadata construct.

.. versionadded:: 1.7.0

.. seealso:: `constructs`, `del_construct`, `has_construct`,
             `set_construct`

:Parameters:

    key: `str`
        The key of the metadata construct.

        *Parameter example:*
          ``key='domainaxis1'``

    default: optional
        Return the value of the *default* parameter if the construct
        does not exist. If set to an `Exception` instance then it will
        be raised instead.

:Returns:

        The metadata construct.

**Examples:**

>>> f.constructs()
{'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degree_N>,
 'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degreeE>,
 'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name:greek_letters(10) >,
 'coordinatereference1': <CoordinateReference: rotated_latitude_longitude>,
 'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>,
 'dimensioncoordinate2': <DimensionCoordinate: grid_longitude(9) degrees>,
 'domainaxis1': <DomainAxis: 10>,
 'domainaxis2': <DomainAxis: 9>}
>>> f.get_construct('dimensioncoordinate1')
<DimensionCoordinate: grid_latitude(10) degrees>

        '''
        return self.constructs.filter_by_key(key).value(default=default)
    #--- End: def

    def has_construct(self, key):
        '''Whather a metadata construct exisits.

.. versionadded:: 1.7.0

.. seealso:: `constructs`, `del_construct`, `get_construct`,
             `set_construct`

:Parameters:

    key: `str`, optional
        The key of the metadata construct.

        *Parameter example:*
          ``key='auxiliarycoordinate0'``

:Returns:

    `bool` 
        True if the construct exisits, otherwise False.

**Examples:**

>>> f.has_construct('dimensioncoordinate1')
True

>>> f.has_construct('dimensioncoordinate99')
False

        '''
        return self.constructs._del_construct(key, default=default)
    #--- End: def

    def set_construct(self, construct, key=None, axes=None,
                      copy=True):
        '''Set a metadata construct.

.. versionadded:: 1.7.0

.. seealso:: `constructs`, `del_construct`, `get_construct`,
             `set_data_axes`

:Parameters:

    construct:
        The metadata construct to be inserted.

    key: `str`, optional
        The construct identifier to be used for the construct. If not
        set then a new, unique identifier is created automatically. If
        the identifier already exisits then the exisiting construct
        will be replaced.

        *Parameter example:*
          ``key='cellmeasure0'``

    axes: (sequence of) `str`, optional
        The construct identifiers of the domain axis constructs
        spanned by the data array. An exception is raised if used for
        a metadata construct that can not have a data array,
        i.e. domain axis, cell method and coordinate reference
        constructs.

        *Parameter example:*
          ``axes='domainaxis1'``
        
        *Parameter example:*
          ``axes=['domainaxis1']``
        
        *Parameter example:*
          ``axes=['domainaxis1', 'domainaxis0']``
        
    copy: `bool`, optional
        If True then set a copy of the construct. By default the
        construct is not copied.
      
:Returns:

     `str`
        The construct identifier for the construct.
    
**Examples:**

>>> key = f.set_construct(c)
>>> key = f.set_construct(c, copy=False)
>>> key = f.set_construct(c, axes='domainaxis2')
>>> key = f.set_construct(c, key='cellmeasure0')

        '''
        return self.constructs._set_construct(construct, key=key,
                                              axes=axes, copy=copy)
    #--- End: def

    def get_data_axes(self, key, default=ValueError):
        '''Return the keys of the domain axis constructs spanned by the data
of a metadata construct.

.. versionadded:: 1.7.0

.. seealso:: `del_data_axes`, `has_data_axes`, `set_data_axes`

:Parameters:

    key: `str`, optional
        Specify a metadata construct.

        *Parameter example:*
          ``key='auxiliarycoordinate0'``

    default: optional
        Return the value of the *default* parameter if the data axes
        have not been set. If set to an `Exception` instance then it
        will be raised instead.

:Returns:

    `tuple` 
        The keys of the domain axis constructs spanned by the data.

**Examples:**

>>> f.set_data_axes(['domainaxis0', 'domainaxis1'])
>>> f.get_data_axes()
('domainaxis0', 'domainaxis1')
>>> f.del_data_axes()
('domainaxis0', 'domainaxis1')
>>> print(f.del_dataxes(None))
None
>>> print(f.get_data_axes(default=None))
None

        '''
        try:
            return self.constructs.data_axes()[key]
        except KeyError:
            return self._default(default, message='2736492783 e28037 TODO')
    #--- End: def

    def del_data_axes(self, key, default=ValueError()):
        '''Remove the keys of the domain axis constructs spanned by the data
of a metadata construct.

.. versionadded:: 1.7.0

.. seealso:: `get_data_axes`, `has_data_axes`, `set_data_axes`

:Parameters:

    key: `str`, optional
        Specify a metadata construct, instead of the field construct.

        *Parameter example:*
          ``key='auxiliarycoordinate0'``

    default: optional
        Return the value of the *default* parameter if the data axes
        have not been set. If set to an `Exception` instance then it
        will be raised instead.

:Returns:

    `tuple` 
        The removed keys of the domain axis constructs spanned by the
        data.

**Examples:**

>>> f.del_data_axes(key='dimensioncoordinate2')
('domainaxis1',)

>>> f.has_data_axes(key='auxiliarycoordinate0')
False
>>> f.has_data_axes(key='auxiliarycoordinate0', default='no axes')
'no axes'

        '''   
        try:
            data_axes = self.constructs.data_axes()[key]
        except KeyError:
            return self._default(default, message='asd aosg8qwg dlb TODO')
        else:
            self.constructs._del_data_axes(key)
                
        return data_axes
    #--- End: def
    
    def has_data_axes(self, key=None):
        '''Whether the domain axis constructs spanned by the data of a
metadata construct have been set.

.. versionadded:: 1.7.0

.. seealso:: `del_data_axes`, `get_data_axes`, `set_data_axes`

:Parameters:

    key: `str`, optional
        Specify a metadata construct.

        *Parameter example:*
          ``key='domainancillary1'``

:Returns:

    `bool` 
        True if domain axis constructs that span the data been set,
        otherwise False.

**Examples:**

>>> f.set_data_axes(['domainaxis0', 'domainaxis1'])
>>> f.get_data_axes()
('domainaxis0', 'domainaxis1')
>>> f.del_data_axes()
('domainaxis0', 'domainaxis1')
>>> print(f.del_dataxes(None))
None
>>> print(f.get_data_axes(default=None))
None

        '''
        axes = self.get_data_axes(key, default=None)
        if axes is None:
            return False

        return True
    #--- End: def
    
    def set_data_axes(self, axes, key):
        '''Set the domain axis constructs spanned by the data of a metadata
construct.

.. versionadded:: 1.7.0

.. seealso:: `data`, `del_data_axes`, `get_data`, `get_data_axes`,
             `has_data_axes`

:Parameters:

     axes: sequence of `str`
        The identifiers of the domain axis constructs spanned by the
        data of the field or of a metadata construct.

        *Parameter example:*
          ``axes='domainaxis1'``

        *Parameter example:*
          ``axes=['domainaxis1']``

        *Parameter example:*
          ``axes=['domainaxis1', 'domainaxis0']``

     key: `str`, optional
        Specify a metadata construct.

        *Parameter example:*
          ``key='domainancillary1'``

:Returns:

    `None`

**Examples:**

>>> f.set_data_axes(['domainaxis0', 'domainaxis1'])
>>> f.get_data_axes()
('domainaxis0', 'domainaxis1')
>>> f.del_data_axes()
('domainaxis0', 'domainaxis1')
>>> print(f.del_dataxes(None))
None
>>> print(f.get_data_axes(None))
None

        '''
        self.constructs._set_construct_data_axes(key=key, axes=axes)
    #--- End: def

#--- End: class



from builtins import super

from .abstract import Implementation

from .. import CF

from .. import (AuxiliaryCoordinate,
                CellMethod,
                CellMeasure,
                CoordinateReference,
                DimensionCoordinate,
                DomainAncillary,
                DomainAxis,
                Field,
                FieldAncillary,
                Bounds,
                InteriorRing,
                CoordinateConversion,
                Datum,
                Count,
                List,
                Index,
                NodeCountProperties,
                PartNodeCountProperties,
)

from ..data import (Data,
                    GatheredArray,
                    NetCDFArray,
                    RaggedContiguousArray,
                    RaggedIndexedArray,
                    RaggedIndexedContiguousArray)


class CFDMImplementation(Implementation):
    '''
    '''
    def __init__(self, cf_version=None,
                 
                 AuxiliaryCoordinate=None,
                 CellMeasure=None,
                 CellMethod=None,
                 CoordinateReference=None,
                 DimensionCoordinate=None,
                 DomainAncillary=None,
                 DomainAxis=None,
                 Field=None,
                 FieldAncillary=None,
                 
                 Bounds=None,
                 InteriorRing=None,
                 CoordinateConversion=None,
                 Datum=None,
                 Data=None,
                 
                 GatheredArray=None,
                 NetCDFArray=None,
                 RaggedContiguousArray=None,
                 RaggedIndexedArray=None,
                 RaggedIndexedContiguousArray=None,
                 
                 List=None,
                 Count=None,
                 Index=None,
                 NodeCountProperties=None,
                 PartNodeCountProperties=None,
    ):
        '''**Initialisation**

    :Parameters:
    
        AuxiliaryCoordinate:
            An auxiliary coordinate construct class.
    
        CellMeasure:
            A cell measure construct class.
    
        CellMethod:
            A cell method construct class.
    
        CoordinateReference:
            A coordinate reference construct class.
    
        DimensionCoordinate:
            A dimension coordinate construct class.
    
        DomainAncillary:
            A domain ancillary construct class.
    
        DomainAxis:
            A domain axis construct class.
    
        Field:
            A field construct class.
    
        FieldAncillary:
            A field ancillary construct class.
    
        Bounds:              = Bounds
        CoordinateAncillary = CoordinateAncillary
        Data                = Data
        NetCDF              = NetCDF

        '''
        super().__init__(
            cf_version=cf_version,

            AuxiliaryCoordinate=AuxiliaryCoordinate,
            CellMeasure=CellMeasure,
            CellMethod=CellMethod,
            CoordinateReference=CoordinateReference,
            DimensionCoordinate=DimensionCoordinate,
            DomainAncillary=DomainAncillary,
            DomainAxis=DomainAxis,
            Field=Field,
            FieldAncillary=FieldAncillary,

            Bounds=Bounds,
            InteriorRing=InteriorRing,
            CoordinateConversion=CoordinateConversion,
            Datum=Datum,
            Data=Data,
            
            GatheredArray=GatheredArray,
            NetCDFArray=NetCDFArray,
            RaggedContiguousArray=RaggedContiguousArray,
            RaggedIndexedArray=RaggedIndexedArray,
            RaggedIndexedContiguousArray=RaggedIndexedContiguousArray,

            List=List,
            Count=Count,
            Index=Index,
            NodeCountProperties=NodeCountProperties,
            PartNodeCountProperties=PartNodeCountProperties,        
        )


    def __repr__(self):
        '''Called by the `repr` built-in function.

    x.__repr__() <==> repr(x)

        '''
        return '<{0}: >'.format(self.__class__.__name__)


    def conform_geometry_variables(self, field):
        '''TODO

    .. versionadded:: 1.8.0

    :Parameters:

        field: Field construct

    :Returns:

        `bool`
            TODO
        '''
        out = {'node_count'     : {},
               'part_node_count': {},
               'interior_ring'  : {}}
        
        for coord in self.get_auxiliary_coordinates(field).values():
            for variable in out:
                x = getattr(self, 'get_'+variable)(coord)
                if x is None:
                    continue
                
                for k, v in x.properties().items():
                    if k not in out[variable]:
                        out[variable][k] = v
                    elif v != out[variable][k]:
                        return False
        #--- End: for
        
        for coord in self.get_auxiliary_coordinates(field).values():
            for variable in out:
                x = getattr(self, 'get_'+variable)(coord)
                if x is None:
                    continue

                x.set_properties(out[variable])
        #--- End: for

        return True

    
    def construct_insert_dimension(self, construct, position):
        '''TODO

    :Parameters:
    
        construct: construct
    
        position: `int`
    
        copy: `bool`, optional
    
    :Returns:
    
        out: construct
    
        '''
        return construct.insert_dimension(position=position)


    def copy_construct(self, construct):
        '''TODO

    :Returns:
    
        out:
            The deep copy.
        '''
        return construct.copy()


    def convert(self, field=None, construct_id=None):
        '''TODO

        '''
        return field.convert(construct_id, full_domain=False)

        
    def del_property(self, construct, prop, default):
        '''TODO

    :Parameters:
    
    :Returns:
    
        out:
            The value of the deleted property
    
    **Examples:**
    
    >>> c.set_properties(f, {'standard_name', 'air_temperature'})
    >>> c.del_property(f, 'standard_name')
    'air_temperature'
    >>> c.get_property(f, 'standard_name')
    AttributeError: Field doesn't have property 'standard_name'
        
        '''
        return construct.del_property(prop, default)


    def field_insert_dimension(self, field, position=0, axis=None):
        '''TODO
        '''
        return field.insert_dimension(axis, position=position)


    def get_array(self, data):
        '''TODO

    :Parameters:

        data: `Data`

        '''
        return data.array


    def get_auxiliary_coordinates(self, field):
        '''TODO
        '''
        return field.auxiliary_coordinates


    def get_bounds(self, parent, default=None):
        '''TODO
        '''
        return parent.get_bounds(default=default)

    
    def get_bounds_ncvar(self, parent, default=None):
        '''TODO
        '''
        bounds = parent.get_bounds(None)
        if bounds is None:
            return default

        return self.nc_get_variable(bounds, default=default)

    
    def get_cell_measures(self, field):
       '''TODO
       '''
       return field.cell_measures


    def get_cell_methods(self, field):
        '''TODO
        '''
        return field.cell_methods.ordered()

       
    def get_cell_method_axes(self, cell_method, default=None):
        '''TODO

    :Returns:
    
        out: `tuple`
        '''
        return cell_method.get_axes(default=default)

    
    def get_cell_method_string(self, cell_method):
        '''TODO

    :Returns:

        out: `str`
        '''
        return str(cell_method)


    def get_cell_method_qualifiers(self, cell_method):
        '''TODO

    :Returns:

        out: `str`
        '''
        return cell_method.qualifiers()


    def get_compressed_array(self, data):
        '''TODO

    :Parameters:
    
        data: `Data`
    
    :Returns:
    
        out: `numpy.ndarray`
        '''
        return data.compressed_array


    def get_compressed_axes(self, field, key=None, construct=None):
        '''TODO

    :Returns:
    
        out: `list`
            '''
        if construct is not None:
            data = self.get_data(construct)
        else:
            data = self.get_data(field)
            
        if key is not None:
            data_axes = self.get_construct_data_axes(field, key)
        else:
            data_axes = self.get_field_data_axes(field)
            
        return [data_axes[i] for i in self.get_data_compressed_axes(data)]


    def get_compression_type(self, construct):
        '''TODO
 
    :Parameters:
    
        construct:      
    
    :Returns:
    
        `str`
            The compression type. If there is no compression then an
            empty string is returned.

        '''
        data = construct.get_data(None)
        if data is None:
            return ''
        
        return data.get_compression_type()

        
    def get_construct_data_axes(self, field, key):
        '''TODO
        '''
        return field.constructs.data_axes()[key]

    
    def get_constructs(self, field, axes=[]):
        '''Return constructs that span particular axes.

    If no axes are specified then all constructs are returned.
    
    If axes are specified then constructs whose data arrays span those
    axes, and possibly other axes, are returned.
    
    :Parameters:
    
        axes: sequence of `str`
    
    :Returns:
    
        `dict`

        '''
        return dict(field.constructs.filter_by_axis('and', *axes))

    
    def get_coordinate_reference_coordinates(self, coordinate_reference):
        '''Return the coordinates of a coordinate reference object.

    :Parameters:
    
        coordinate_reference: `CoordinateReference`
    
    :Returns:
    
        out: `set`

        '''
        return coordinate_reference.coordinates()

    
    def get_coordinate_conversion_parameters(self, coordinate_reference):
        '''TODO

    :Returns:

        out: `dict`

        '''
        return coordinate_reference.coordinate_conversion.parameters()


    def get_coordinate_references(self, field):
        '''TODO
        '''
        return field.coordinate_references


    def get_coordinates(self, field):
       '''TODO

    :Parameters:

       '''
       return field.coordinates

    
    def get_data_calendar(self, data, default=None):
        '''TODO
        '''
        return data.get_calendar(default=default)


    def get_data_compressed_axes(self, data):
        '''TODO

    :Returns:

        out: `list`
        '''
        return data.get_compressed_axes()


    def get_data_ndim(self, parent):
        '''Return the number of dimensions spanned by the data array.

    :Parameters:
    
        parent: 
            The object containing the data array.
    
    :Returns:
    
        out: `int`
            The number of dimensions spanned by the data array.
    
    **Examples:**
    
    >>> d
    <DimensionCoordinate: latitude(180) degrees_north>
    >>> w.get_data_ndim(d)
    1
    
    >>> b
    <Bounds: latitude(180, 2) degrees_north>
    >>> w.get_data_ndim(b)
    2
        '''
        return parent.data.ndim


    def get_data_size(self, parent):
        '''Return the number of elements in the data array.

:Parameters:

    parent: 
        The object containing the data array.

:Returns:

    out: `int`
        The number of elements in the data array.

**Examples:**

>>> d
<DimensionCoordinate: latitude(180) degrees_north>
>>> w.get_data_ndim(d)
180

>>> b
<Bounds: latitude(180, 2) degrees_north>
>>> w.get_data_ndim(b)
360
        '''
        return parent.data.size
    #--- End: def

    def get_data_units(self, data, default=None):
        '''
        '''
        return data.get_units(default=default)
    #--- End: def

    def get_datum(self, coordinate_reference):
        '''
        
:Parameters:

:Returns:

    out: datum object
        
        '''
        return coordinate_reference.datum
    #--- End: def
            
    def get_datum_parameters(self, ref):
        '''Return the parameter-valued terms of a coordinate reference datum.

:Parameters:

    coordinate_reference: `CoordinateReference`

:Returns:

    out: `dict`

        '''        
        return ref.datum.parameters()
    #--- End: def

    def get_dimension_coordinates(self, field):
        '''
        '''
        return field.dimension_coordinates
    #--- End: def

    def get_domain_ancillaries(self, field):
       '''
:Parameters:

       '''
       return field.domain_ancillaries
    #--- End: def

    def get_domain_axes(self, field):
        '''
        '''
        return field.domain_axes
    #--- End: def
 
    def get_domain_axis_size(self, field, axis):
        '''
        '''
        return field.domain_axes[axis].get_size()
    #--- End: def

    def get_sample_dimension_position(self, construct):
        '''
        '''
        return construct.get_data().get_compressed_dimension()
    #--- End: def

#    def nc_get_instance_dimension(self, index, default=None):
#        '''Return the name of the netCDF instance dimension.
#
#:Returns: 
#
#    `str`
#        The name of the netCDF instance dimension.
#
#        '''
#        return index.nc_get_instance_dimension(default=default)
#    #--- End: def

    def nc_get_geometry_variable(self, field, default=None):
        '''TODO

.. versionadded:: 1.8.0

:Parameters:

:Returns:

    `str`

        '''
        return field.nc_get_geometry_variable(default)
    #--- End: def

    def nc_get_hdf5_chunksizes(self, data):
        '''Return the HDF5 chunksizes for the data. 

..versionadded:: 1.7.2

:Parameters:

    data: Data instance

:Returns: 

    `tuple` or `None`
        The HDF5 chunksizes, or `None` if they haven't been set.

        '''
        out = data.nc_hdf5_chunksizes()
        if not out:
            out = None

        return out
    #--- End: def

    def nc_get_sample_dimension(self, count, default=None):
        '''Return the name of the netCDF sample dimension.

:Returns: 

    `str`
        The name of the netCDF sample dimension.

        '''
        return count.nc_get_sample_dimension(default=default)
    #--- End: def

    def nc_is_unlimited_axis(self, field, axis):
        '''Whether a domain axis corresponds to a netCDF unlimited dimension.

.. versionadded:: 1.7.0

:Parameters:
  
    field: `Field`

:Returns:

    out: `set`
        The selection of domain axis construct identifiers that are
        unlimited.

        '''
        domain_axis = field.constructs.get(axis)
        if domain_axis is None:
            return False

        return domain_axis.nc_is_unlimited()
    #--- End: def
    
    def nc_set_unlimited_axis(self, field, axis):
        '''Set a domain axis to correspond to a netCDF unlimited dimension.

.. versionadded:: 1.7.4

:Parameters:
  
    field: `Field`

    axis: `str`
        Domain axis construct key.

:Returns:

    `None`

        '''
        domain_axis = field.constructs.get(axis)
        if domain_axis is None:
            return

        domain_axis.nc_set_unlimited(True)
    #--- End: def
    
#    def nc_set_unlimited_dimensions(self, field, unlimited):
#        '''Set the selection of domain axis constructs that were read in as
#netCDF unlimited dimensions.
#
#.. versionadded:: 1.7.0
#
#:Parameters:
#  
#    field: `Field`
#
#    unlimited: sequence of `str`
#
#:Returns:
#
#    out: `set`
#        The selection of domain axis construct identifiers that are
#        unlimited.
#
#        '''
#        return field.nc_set_unlimited_dimensions(unlimited)
#    #--- End: def
    
    def nc_get_global_attributes(self, field):
        '''TODO
        '''
        return field.nc_global_attributes()
    #--- End: def
    
    def nc_set_global_attributes(self, field, attributes):
        '''TODO

.. versionadded:: 1.7.0

attributes: `dict`

        '''
        for attr, value in attributes.items():
            field.nc_set_global_attribute(attr, value)
    #--- End: def
    
    def equal_constructs(self, construct0, construct1,
                         ignore_type=False):
        '''

        '''    
        return construct0.equals(construct1, ignore_type=ignore_type)
    #--- End: def

    def equal_properties(self, property_value0, property_value1):
        '''

        '''
        field = self.get_class('Field')()
        return field._equals(property_value0, property_value1)
    #--- End: def

    def equal_datums(self, coordinate_reference0, coordinate_reference1):
        ''':Parameters:

    coordinate_reference0: coordinate reference construct

    coordinate_reference1: coordinate reference construct

:Returns:

    out: `bool`

        '''
        datum0 = coordinate_reference0.datum
        datum1 = coordinate_reference1.datum
        return datum0.equals(datum1)
    #--- End: def

    def get_construct_data_size(self, construct):
        '''

:Parameters:

:Returns:

    out: `int`

        '''
        return construct.data.size
    #--- End: def
   
    def nc_get_external(self, parent):
        '''Return whether a construct is external.

:Parameters:

    parent: 
        The object

:Returns:

    out: `bool`
        Whether the construct is external.

**Examples:**
        '''
        if not hasattr(parent, 'nc_get_external'):
            return False
        
        return parent.nc_get_external()
    #--- End: def
    
    def get_field_ancillaries(self, field):
        '''Return the field ancillaries of a field.

:Parameters:

    field: 
        The field object.

:Returns:

    out: `dict`
        A dictionary whose values are field ancillary objects, keyed
        by unique identifiers.

**Examples:**

>>> w.get_field_ancillaries(f)
{'fieldancillary0': <FieldAncillary: ....>,
 'fieldancillary1': <FieldAncillary: ....>}

        '''
        return field.field_ancillaries
    #--- End: def
                  
    def get_field_data_axes(self, field):
        '''
        '''
        return field.get_data_axes()
    #--- End: def
         
    def get_data_max(self, parent):
        '''

:Parameters:

:Returns:

    out: `int`
        
        '''
        return parent.data.max()
    #--- End: def
    
    def get_data_sum(self, parent):
        '''

:Parameters:

:Returns:

    out: `int`
        
        '''
        return parent.data.sum()
    #--- End: def
    
    def get_count(self, construct):
        '''Return the measure property of a cell measure contruct.

:Parameters:

    cell_measure:
        The cell measure object.

:Returns:

    out: `str` or `None`
        The measure property, or `None` if it has not been set.

**Examples:**

>>> c
<CellMeasure: area(73, 96) km2>
>>> w.get_measure(c)
'area'
        '''
        return construct.get_data().get_count(default=None)
    #--- End: def
    
    def get_index(self, construct):
        '''TODO

    :Parameters:
    
      
    :Returns:
    
        out: 

        '''
        return construct.get_data().get_index(default=None)

    
    def get_interior_ring(self, construct):
        '''TODO

    :Parameters:
    
        TODO

    :Returns:
    
        TODO
    
        '''
        return construct.get_interior_ring(default=None)

    
    def get_list(self, construct):
        '''TODO

        '''
        return construct.get_data().get_list(default=None)

    
    def get_measure(self, cell_measure):
        '''Return the measure property of a cell measure contruct.

:Parameters:

    cell_measure:
        The cell measure object.

:Returns:

    out: `str` or `None`
        The measure property, or `None` if it has not been set.

**Examples:**

>>> c
<CellMeasure: area(73, 96) km2>
>>> w.get_measure(c)
'area'
        '''
        return cell_measure.get_measure(default=None)
    #--- End: def
    
    def nc_get_dimension(self, parent, default=None):
        '''Return the netCDF variable name.

:Parameters:

    parent: 
        The object containing the data array.

    default: `str`, optional

:Returns:

    out: 
        The netCDF dimension name.

**Examples:**
        '''
        return parent.nc_get_dimension(default=default)
    #--- End: def

    def nc_get_variable(self, construct, default=None):
       '''TODO

    :Parameters:
    
       TODO

    :Returns:
    
       `str`
           TODO
       '''
       return construct.nc_get_variable(default=default)


    def get_node_count(self, construct):
        '''TODO

    .. versionadded:: 1.8.0
    
    :Parameters:
    
        construct: 
      
    :Returns:
    
        Node count variable

        '''
        return construct.get_node_count(default=None)

    
    def get_part_node_count(self, construct):
        '''TODO

    .. versionadded:: 1.8.0
    
    :Parameters:
    
      
    :Returns:
    
        Part node count variable
    

        '''
        return construct.get_part_node_count(default=None)

    
    def get_properties(self, parent):
        '''Return all properties.

:Parameters:

    parent: 
        The object containing the properties.

:Returns:

    out: `dict`
        The property names and their values

**Examples:**

>>> d
<DimensionCoordinate: latitude(180) degrees_north>
>>> w.get_properties(d)
{'units: 'degrees_north'}
 'standard_name: 'latitude',
 'foo': 'bar'}
        '''
        return parent.properties()
    #--- End: def

    def get_property(self, construct, prop, default=None):
       '''TODO

    :Parameters:

    :Returns:

       out:
       '''
       return construct.get_property(prop, default=default)


    def get_geometry(self, construct, default=None):
       '''TODO

    :Parameters:

    :Returns:

       out:
       '''
       return construct.get_geometry(default=default)


    def get_data(self, parent, default=None):
        '''Return the data array.

    :Parameters:
    
        parent: 
            The object containing the data array.
    
    :Returns:
    
            The data.
    
    **Examples:**
    
    >>> d
    <DimensionCoordinate: latitude(180) degrees_north>
    >>> w.get_data(d)
    <Data(180): [-89.5, ..., 89.5] degrees_north>
    
    >>> b
    <Bounds: latitude(180, 2) degrees_north>
    >>> w.get_data(b)
    <Data(180, 2): [[-90, ..., 90]] degrees_north>
        '''
        return parent.get_data(default=default)


    def initialise_AuxiliaryCoordinate(self):
        '''TODO
        '''
        cls = self.get_class('AuxiliaryCoordinate')
        return cls()


    def initialise_Bounds(self):
        '''TODO
        '''
        cls = self.get_class('Bounds')
        return cls()


    def initialise_CellMeasure(self, measure=None):
        '''TODO
        '''
        cls = self.get_class('CellMeasure')
        return cls(measure=measure)


    def initialise_CellMethod(self, axes=None, method=None,
                              qualifiers=None):
        '''TODO
        '''
        cls = self.get_class('CellMethod')
        return cls(axes=axes, method=method, qualifiers=qualifiers)


    def initialise_CoordinateConversion(self, 
                                        domain_ancillaries=None,
                                        parameters=None):
        '''TODO
        '''
        cls = self.get_class('CoordinateConversion')
        return cls(domain_ancillaries=domain_ancillaries,
                   parameters=parameters)


    def initialise_CoordinateReference(self):
        '''TODO
        '''
        cls = self.get_class('CoordinateReference')
        return cls()

 
    def initialise_Count(self):
        '''TODO

        '''
        cls = self.get_class('Count')
        return cls()


    def initialise_Data(self, array=None, units=None, calendar=None,
                        copy=True, **kwargs):
        '''TODO

    :Patameters:
    
        array:
    
        units:
    
        calendar:
    
        copy: `bool`, optional
    
        kwargs: optional
            Not used in this imlementation
    
    :Returns;
    
        `Data`
            TODO

        '''
        cls = self.get_class('Data')
        return cls(array=array, units=units, calendar=calendar,
                   copy=copy, **kwargs)


    def initialise_Datum(self, parameters=None):
        '''TODO
        '''
        cls = self.get_class('Datum')
        return cls(parameters=parameters)


    def initialise_DimensionCoordinate(self, properties=None,
                                       data=None, bounds=None,
                                       interior_ring=None, copy=True):
        '''TODO
        '''
        cls = self.get_class('DimensionCoordinate')
        return cls(properties=properties, data=data, bounds=bounds,
                   interior_ring=interior_ring, copy=copy)


    def initialise_DimensionCoordinate_from_AuxiliaryCoordinate(            
            self,
            auxiliary_coordinate=None,
            copy=True):
        '''TODO
        '''
        cls = self.get_class('DimensionCoordinate')
        return cls(source=auxiliary_coordinate, copy=copy)


    def initialise_DomainAncillary(self):
        '''TODO
        '''
        cls = self.get_class('DomainAncillary')
        return cls()


    def initialise_DomainAxis(self, size=None):
        '''TODO
        '''
        cls = self.get_class('DomainAxis')
        return cls(size=size)


    def initialise_Field(self):
        '''TODO
        '''
        cls = self.get_class('Field')
        return cls()


    def initialise_FieldAncillary(self):
        '''TODO
        '''
        cls = self.get_class('FieldAncillary')
        return cls()


    def initialise_GatheredArray(self, compressed_array=None,
                                 ndim=None, shape=None, size=None,
                                 compressed_dimension=None,
                                 list_variable=None):
        '''TODO

        '''
        cls = self.get_class('GatheredArray')
        return cls(compressed_array=compressed_array, ndim=ndim,
                   shape=shape, size=size,
                   compressed_dimension=compressed_dimension,
                   list_variable=list_variable)


    def initialise_Index(self):
        '''TODO

        '''
        cls = self.get_class('Index')
        return cls()


    def initialise_InteriorRing(self):
        '''TODO

        '''
        cls = self.get_class('InteriorRing')
        return cls()


    def initialise_List(self):
        '''TODO

        '''
        cls = self.get_class('List')
        return cls()


    def initialise_NetCDFArray(self, filename=None, ncvar=None,
                               dtype=None, ndim=None, shape=None,
                               size=None):
        '''TODO

        '''
        cls = self.get_class('NetCDFArray')
        return cls(filename=filename, ncvar=ncvar, dtype=dtype,
                   ndim=ndim, shape=shape, size=size)


    def initialise_NodeCount(self):
        '''

        '''
        cls = self.get_class('NodeCountProperties')
        return cls()


    def initialise_PartNodeCount(self):
        '''TODO

        '''
        cls = self.get_class('PartNodeCountProperties')
        return cls()


    def initialise_RaggedContiguousArray(self, compressed_array=None,
                                         ndim=None, shape=None,
                                         size=None,
                                         count_variable=None):
        '''TODO
        '''
        cls = self.get_class('RaggedContiguousArray')
        return cls(compressed_array=compressed_array, ndim=ndim,
                   shape=shape, size=size,
                   count_variable=count_variable)


    def initialise_RaggedIndexedArray(self, compressed_array=None,
                                      ndim=None, shape=None,
                                      size=None, index_variable=None):
        '''TODO
        '''
        cls = self.get_class('RaggedIndexedArray')
        return cls(compressed_array=compressed_array, ndim=ndim,
                   shape=shape, size=size,
                   index_variable=index_variable)


    def initialise_RaggedIndexedContiguousArray(self,
                                                compressed_array=None,
                                                ndim=None, shape=None,
                                                size=None,
                                                count_variable=None,
                                                index_variable=None):
        '''TODO
        '''
        cls = self.get_class('RaggedIndexedContiguousArray')
        return cls(compressed_array=compressed_array, ndim=ndim,
                   shape=shape, size=size,
                   count_variable=count_variable,
                   index_variable=index_variable)


    def is_climatology(self, coordinate):
        '''TODO

    :Returns:
    
        out: `bool`
            The value of the 'climatology' cell extent parameter, or
            False if not set.

        '''
        return bool(coordinate.get_geometry(None) == 'climatology')


    def is_geometry(self, coordinate):
        '''Return True if the coordinate bounds are geometries.

    :Parameters:
    
        coordinate: 
            The coordinate construct.
    
    :Returns:
    
        out: `bool`
             True if the coordinate bounds are geometries, otherwise
             False.

        '''
        return bool(coordinate.get_geometry(None) in ('point', 'line', 'polygon'))


    def nc_set_instance_dimension(self, variable, ncdim):
        '''TODO
        '''
        variable.nc_set_instance_dimension(ncdim)

    

    def nc_set_sample_dimension(self, variable, ncdim):
        '''TODO
        '''
        variable.nc_set_sample_dimension(ncdim)

    
    def set_auxiliary_coordinate(self, field, construct, axes, copy=True):
        '''Insert a auxiliary coordinate object into a field.
        
    :Parameters:
    
        field: `Field`
    
        construct: `AuxiliaryCoordinate`
    
        axes: `tuple`
    
        copy: `bool`, optional
    
    :Returns:
    
        out: `str`
        '''
        return field.set_construct(construct, axes=axes, copy=copy)


    def set_bounds(self, construct, bounds, copy=True):
        '''TODO

    :Parameters:
    
    :Returns:
    
        `None`

        '''
        construct.set_bounds(bounds, copy=copy)


    def set_cell_measure(self, field, construct, axes, copy=True):
        '''Insert a cell_measure object into a field.

    :Parameters:
    
        field: `Field`
    
        construct: `CellMeasure`
    
        axes: `tuple`
    
        copy: `bool`, optional
    
    :Returns:
    
        out: `str`
        '''
        return field.set_construct(construct, axes=axes, copy=copy)


    def set_cell_method(self, field, construct, copy=True):
        '''Insert a cell_method object into a field.

    :Parameters:
    
        field: `Field`
    
        construct: `CellMethod`
    
        copy: `bool`, optional
    
    :Returns:
    
        out: `str`
        '''
        return field.set_construct(construct, copy=copy)


    def set_cell_method_axes(self, cell_method, axes):
        '''TODO
        '''
        cell_method.set_axes(axes)

    
    def set_cell_method_method(self, cell_method, method):
        '''TODO
        '''
        cell_method.set_method(method)

    
    def set_coordinate_conversion(self, coordinate_reference, coordinate_conversion):
        '''TODO

    :Parameters:
    
    :Returns:
    
        `None`
    
        '''
        coordinate_reference.set_coordinate_conversion(coordinate_conversion)


    def set_coordinate_reference(self, field, construct, copy=True):
        '''Insert a coordinate reference object into a field.

    :Parameters:
        field: `Field`

        construct: `CoordinateReference`

        copy: `bool`, optional

    :Returns:

        out: `str`
            '''
        return field.set_construct(construct, copy=copy)


    def set_coordinate_reference_coordinates(self,
                                             coordinate_reference,
                                             coordinates):
        '''TODO

    :Parameters:
    
    :Returns:
    
        `None`
        '''
        coordinate_reference.set_coordinates(coordinates)


    def set_coordinate_reference_coordinate(self,
                                            coordinate_reference,
                                            coordinate):
        '''TODO
        '''
        coordinate_reference.set_coordinate(coordinate)


    def set_data(self, construct, data, axes=None, copy=True):
        '''If the construct is a Field then the corresponding domain axes must
    also be provided.
    
    :Parameters:
    
        construct: construct
    
        data: `Data`
    
        axes: `tuple`, optional
    
        copy: `bool`, optional
    
    :Returns:
    
        `None`

        '''
        if axes is None:
            construct.set_data(data, copy=copy)
        else:
            construct.set_data(data, axes=axes, copy=copy)
    #--- End: def

    def set_datum(self, coordinate_reference, datum):
        '''

    :Parameters:
    
    :Returns:
    
        `None`

        '''
        coordinate_reference.set_datum(datum)
    #--- End: def

    def set_dimension_coordinate(self, field, construct, axes, copy=True):
        '''Insert a dimension coordinate object into a field.
        
    :Parameters:
    
        field: `Field`
    
        construct: `DimensionCoordinate`
    
        axes: `tuple`
    
        copy: `bool`, optional
    
    :Returns:
    
        out: `str`
        '''
        return field.set_construct(construct, axes=axes, copy=copy)
    #--- End: def

    def set_domain_ancillary(self, field, construct, axes, copy=True):
#                             extra_axes=0, copy=True):
        '''Insert a domain ancillary object into a field.

    :Parameters:
    
        field: `Field`
    
        construct: `DomainAncillary`
    
        axes: `tuple`
    
        copy: `bool`, optional
    
    :Returns:
    
        out: `str`
        '''
        return field.set_construct(construct, axes=axes, copy=copy)
    #--- End: def

    def set_domain_axis(self, field, construct, copy=True):
        '''Insert a domain_axis object into a field.

    :Parameters:
    
        field: `Field`
    
        construct: `domain_axis`
    
        copy: `bool`, optional
    
    :Returns:
    
        out: `str`
        '''
        return field.set_construct(construct, copy=copy)
    #--- End: def

    def nc_set_external(self, construct):
        '''TODO
        '''
        construct.nc_set_external(True)
    #--- End: def

    def set_field_ancillary(self, field, construct, axes, copy=True):
        '''Insert a field ancillary object into a field.

    :Parameters:
    
        field: `Field`
    
        construct: `FieldAncillary`
    
        axes: `tuple`
    
        copy: `bool`, optional
    
    :Returns:
    
        out: `str`
        '''
        return field.set_construct(construct, axes=axes, copy=copy)
    #--- End: def

    def set_geometry(self, coordinate, value):
        '''TODO

    .. versionadded:: 1.8.0

        '''
        coordinate.set_geometry(value)


    def set_node_count(self, parent, node_count, copy=True):
        '''Insert TODO

    Parameters:
    
       copy: `bool`, optional
    
    Returns:
    
       `None`
        '''
        parent.set_node_count(node_count, copy=copy)


    def set_part_node_count(self, parent, part_node_count, copy=True):
        '''Insert TODO

    .. versionadded:: 1.8.0
    
    :Parameters:
    
        copy: `bool`, optional
    
    :Returns:
    
        `None`
        '''
        parent.set_part_node_count(part_node_count, copy=copy)


    def set_interior_ring(self, parent, interior_ring, copy=True):
        '''Insert an interior ring array into a coordiante.

    .. versionadded:: 1.8.0
    
    :Parameters:
    
        copy: `bool`, optional
    
    :Returns:
    
        `None`

        '''
        parent.set_interior_ring(interior_ring, copy=copy)


    def set_dataset_compliance(self, field, report):
        '''TODO

    ..versionadded:: 1.7

        '''
        field._set_dataset_compliance(report)

    
    def nc_set_dimension(self, construct, ncdim):
        '''TODO

    :Parameters:

    :Returns:

        `None`

        '''
        construct.nc_set_dimension(ncdim)


    def nc_set_geometry_variable(self, field, ncvar):
        '''TODO

    :Parameters:

    :Returns:

        `None`

        '''
        field.nc_set_geometry_variable(ncvar)

    
    def nc_set_variable(self, parent, ncvar):
        '''TODO

    :Parameters:
    
    :Returns:

        `None`
        '''
        parent.nc_set_variable(ncvar)


    def nc_get_datum_variable(self, ref):
        '''TODO

    .. versionadded:: 1.7.5

    :Parameters:

    :Returns:

        `str` or `None`
        '''
        return ref.nc_get_datum_variable(default=None)


    def nc_set_datum_variable(self, ref, ncvar):
        '''TODO

    .. versionadded:: 1.7.5

    :Parameters:

    :Returns:

        `None`
        '''
        ref.nc_set_datum_variable(ncvar)


    def set_properties(self, construct, properties, copy=True):
        '''TODO

    :Parameters:

    :Returns:

        `None`

        '''
        construct.set_properties(properties, copy=copy)

 
    def has_bounds(self, construct):
        '''
:Parameters:

:Returns:

    out: `bool`
        '''
        return construct.has_bounds()
    #--- End: def
    
    
    def has_datum(self, coordinate_reference):
        '''Return True if a coordinate reference has a datum.

    :Parameters:
    
        coordinate_reference: coordinate reference construct
    
    :Returns:
    
        out: `bool`
    
    **Examples:**
    
    >>> if API.has_datum(ref):
    ...     print ref, 'has a datum'
    ... else:
    ...     print ref, 'does not have a datum'
    
        '''
        return bool(coordinate_reference.datum)

    
    def has_property(self, parent, prop):
        '''Return True if a property exists.

    :Parameters:
    
        parent: 
            The object containing the property.
    
    :Returns:
    
        out: `bool`
            `True` if the property exists, otherwise `False`.
    
    **Examples:**
    
    >>> coord
    <DimensionCoordinate: latitude(180) degrees_north>
    >>> w.has_property(coord, 'units')
    True
    
    >>> bounds
    <Bounds: latitude(180, 2) degrees_north>
    >>> w.has_property(bounds, 'long_name')
    False
        '''
        return parent.has_property(prop)

    
    def squeeze(self, construct, axes=None):
        '''TODO
        '''
        return construct.squeeze(axes=axes)

        
#--- End: class


_implementation = CFDMImplementation(
    cf_version = CF(),
    
    AuxiliaryCoordinate = AuxiliaryCoordinate,
    CellMeasure         = CellMeasure,
    CellMethod          = CellMethod,
    CoordinateReference = CoordinateReference,
    DimensionCoordinate = DimensionCoordinate,
    DomainAncillary     = DomainAncillary,
    DomainAxis          = DomainAxis,
    Field               = Field,
    FieldAncillary      = FieldAncillary,
    
    Bounds       = Bounds,
    InteriorRing = InteriorRing,
    
    CoordinateConversion = CoordinateConversion,
    Datum                = Datum,

    List                    = List,
    Index                   = Index,
    Count                   = Count,
    NodeCountProperties     = NodeCountProperties,
    PartNodeCountProperties = PartNodeCountProperties,
    
    Data                         = Data,
    GatheredArray                = GatheredArray,
    NetCDFArray                  = NetCDFArray,
    RaggedContiguousArray        = RaggedContiguousArray,
    RaggedIndexedArray           = RaggedIndexedArray,
    RaggedIndexedContiguousArray = RaggedIndexedContiguousArray,
)

def implementation():
    '''Return a container for the CF data model implementation.

    .. versionadded:: 1.7.0
    
    .. seealso:: `cfdm.read`, `cfdm.write`
    
    :Returns:
    
        `CFDMImplementation`
            A container for the CF data model implementation.
    
    **Examples:**
    
    >>> i = cfdm.implementation()
    >>> i
    <CFDMImplementation: >
    >>> i.classes()
    {'AuxiliaryCoordinate': cfdm.auxiliarycoordinate.AuxiliaryCoordinate,
     'Bounds': cfdm.bounds.Bounds,
     'CellMeasure': cfdm.cellmeasure.CellMeasure,
     'CellMethod': cfdm.cellmethod.CellMethod,
     'CoordinateConversion': cfdm.coordinateconversion.CoordinateConversion,
     'CoordinateReference': cfdm.coordinatereference.CoordinateReference,
     'Count': cfdm.count.Count,
     'Data': cfdm.data.data.Data,
     'Datum': cfdm.datum.Datum,
     'DimensionCoordinate': cfdm.dimensioncoordinate.DimensionCoordinate,
     'DomainAncillary': cfdm.domainancillary.DomainAncillary,
     'DomainAxis': cfdm.domainaxis.DomainAxis,
     'Field': cfdm.field.Field,
     'FieldAncillary': cfdm.fieldancillary.FieldAncillary,
     'GatheredArray': cfdm.data.gatheredarray.GatheredArray,
     'Index': cfdm.index.Index,
     'InteriorRing': cfdm.interiorring.InteriorRing,
     'List': cfdm.list.List,
     'NetCDFArray': cfdm.data.netcdfarray.NetCDFArray,
     'NodeCountProperties': cfdm.nodecount.NodeCountProperties,
     'PartNodeCountProperties': cfdm.partnodecount.PartNodeCountProperties,
     'RaggedContiguousArray': cfdm.data.raggedcontiguousarray.RaggedContiguousArray,
     'RaggedIndexedArray': cfdm.data.raggedindexedarray.RaggedIndexedArray,
     'RaggedIndexedContiguousArray': cfdm.data.raggedindexedcontiguousarray.RaggedIndexedContiguousArray}

    '''
    return _implementation.copy()




from builtins import super
from past.builtins import basestring

from . import abstract
from . import mixin

from . import Constructs


class Field(mixin.ConstructAccess, abstract.PropertiesData):
    '''A field construct of the CF data model.

The field construct is central to the CF data model, and includes all
the other constructs. A field corresponds to a CF-netCDF data variable
with all of its metadata. All CF-netCDF elements are mapped to a field
construct or some element of the CF field construct. The field
construct contains all the data and metadata which can be extracted
from the file using the CF conventions.

The field construct consists of a data array and the definition of its
domain (that describes the locations of each cell of the data array),
field ancillary constructs containing metadata defined over the same
domain, and cell method constructs to describe how the cell values
represent the variation of the physical quantity within the cells of
the domain. The domain is defined collectively by the following
constructs of the CF data model: domain axis, dimension coordinate,
auxiliary coordinate, cell measure, coordinate reference and domain
ancillary constructs. All of the constructs contained by the field
construct are optional.

The field construct also has optional properties to describe aspects
of the data that are independent of the domain. These correspond to
some netCDF attributes of variables (e.g. units, long_name and
standard_name), and some netCDF global file attributes (e.g. history
and institution).

.. versionadded:: 1.7.0

    '''
    # ----------------------------------------------------------------
    # Define the base of the identity keys for each construct type
    # ----------------------------------------------------------------
    _construct_key_base = {'auxiliary_coordinate': 'auxiliarycoordinate',
                           'cell_measure'        : 'cellmeasure',
                           'cell_method'         : 'cellmethod',
                           'coordinate_reference': 'coordinatereference',
                           'dimension_coordinate': 'dimensioncoordinate',
                           'domain_ancillary'    : 'domainancillary',
                           'domain_axis'         : 'domainaxis',
                           'field_ancillary'     : 'fieldancillary',
    }
    
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._Constructs = Constructs
        return instance
    #--- End: def

    def __init__(self, properties={}, source=None, copy=True,
                 _use_data=True):
        '''**Initialization**
        
:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
           ``properties={'standard_name': 'air_temperature'}``
        
        Properties may also be set after initialisation with the
        `set_properties` and `set_property` methods.

    source: optional
        Initialize the properties, data and metadata constructs from
        those of *source*.
        
    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(properties=properties, source=source,
                         copy=copy, _use_data=False)

        if source is not None:
            # Initialise constructs and the data from the source
            # parameter
            try:
                constructs = source.constructs
            except AttributeError:
                constructs = None
                
            try:
                data = source.get_data(default=None) 
            except AttributeError:
                data = None

            try:
                data_axes = source.get_data_axes(default=None)
            except AttributeError:
                data_axes = None
                                
            if constructs is not None and (copy or not _use_data):
                constructs = constructs.copy(data=_use_data)
        else:
            constructs = None
            data       = None
            data_axes  = None

        if constructs is None:
            constructs = self._Constructs(**self._construct_key_base)
            
        self._set_component('constructs', constructs, copy=False)

        if data is not None and _use_data:
            self.set_data(data, data_axes, copy=copy)
        elif data_axes is not None:
            self.set_data_axes(data_axes)
     #--- End: def
    
    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def construct_type(self):
        '''Return a description of the construct type.
        
.. versionadded:: 1.7.0
        
:Returns:

    `str`
        The construct type.

**Examples:**

>>> f.construct_type
'field'

        '''
        return 'field'
    #--- End: def

    @property
    def constructs(self):
        '''Return the metdata constructs.

.. versionadded:: 1.7.0

:Returns:

    `Constructs`
        The constructs.

**Examples:**

>>> print(f.constructs)
Constructs:
{'cellmethod0': <CellMethod: area: mean>,
 'dimensioncoordinate0': <DimensionCoordinate: latitude(5) degrees_north>,
 'dimensioncoordinate1': <DimensionCoordinate: longitude(8) degrees_east>,
 'dimensioncoordinate2': <DimensionCoordinate: time(1) days since 2018-12-01 >,
 'domainaxis0': <DomainAxis: size(5)>,
 'domainaxis1': <DomainAxis: size(8)>,
 'domainaxis2': <DomainAxis: size(1)>}

        '''
        return self._get_component('constructs')
    #--- End: def

    @property
    def domain(self):
        '''Return the domain.
        
``f.domain`` is equivalent to ``f.get_domain()`` 

.. versionadded:: 1.7.0

.. seealso:: `get_domain`

:Returns:

    `Domain`
        The domain.

**Examples:**

>>> d0 = f.domain
>>> d1 = f.get_domain()
>>> d0.equals(d1)
True
 
        '''
        return self.get_domain()
    #--- End: def
    
    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def del_data_axes(self, key=None, default=ValueError()):
        '''Remove the keys of the domain axis constructs spanned by the data
of the field or of a metadata construct.

.. versionadded:: 1.7.0

.. seealso:: `get_data_axes`, `has_data_axes`, `set_data_axes`

:Parameters:

    key: `str`, optional
        Specify a metadata construct, instead of the field construct.

        *Parameter example:*
          ``key='auxiliarycoordinate0'``

    default: optional
        Return the value of the *default* parameter if the data axes
        have not been set. If set to an `Exception` instance then it
        will be raised instead.

:Returns:

    `tuple` 
        The removed keys of the domain axis constructs spanned by the
        data.

**Examples:**

>>> f.del_data_axes()
('domainaxis0', 'domainaxis1')

>>> f.del_data_axes(key='dimensioncoordinate2')
('domainaxis1',)

>>> f.has_data_axes()
False
>>> f.has_data_axes(default='no axes')
'no axes'

        '''   
        if key is not None:
            return super().del_data_axes(key, default=default)

        try:
            return self._del_component('data_axes')
        except ValueError:
            return self._default(default,
              "{!r} has no data axes".format(self.__class__.__name__))
    #--- End: def

    def get_domain(self):
        '''Return the domain.

.. versionadded:: 1.7.0

.. seealso:: `domain`

:Returns:

    `Domain`
         The domain.

**Examples:**

>>> d = f.get_domain()

        '''
        return self._Domain.fromconstructs(self.constructs)
    #--- End: def

    def get_data_axes(self, key=None, default=ValueError()):
        '''Return the keys of the domain axis constructs spanned by the data
of the field or of a metadata construct.

.. versionadded:: 1.7.0

.. seealso:: `del_data_axes`, `get_data`, `set_data_axes`

:Parameters:

    key: `str`, optional
        Specify a metadata construct, instead of the field construct.

        *Parameter example:*
          ``key='auxiliarycoordinate0'``

    default: optional
        Return the value of the *default* parameter if the data axes
        have not been set. If set to an `Exception` instance then it
        will be raised instead.

:Returns:

    `tuple` 
        The keys of the domain axis constructs spanned by the data.

**Examples:**

>>> f.get_data_axes()
('domainaxis0', 'domainaxis1')

>>> f.get_data_axes('dimensioncoordinate2')
('domainaxis1',)

>>> f.has_data_axes()
False
>>> f.get_data_axes(default='no axes')
'no axes'

        '''
        if key is not None:
            return super().get_data_axes(key, default=default)
        
        try:
            return self._get_component('data_axes')
        except ValueError:
            return self._default(default,
              "{!r} has no data axes".format(self.__class__.__name__))
    #--- End: def

    def has_data_axes(self, key=None):
        '''Whether the domain axis constructs spanned by the data of the field
or of a metadata construct have been set.

.. versionadded:: 1.7.0

.. seealso:: `del_data_axes`, `get_data_axes`, `set_data_axes`

:Parameters:

    key: `str`, optional
        Specify a metadata construct, instead of the field construct.

        *Parameter example:*
          ``key='domainancillary1'``

:Returns:

    `bool` 
        True if domain axis constructs that span the data been set,
        otherwise False.

**Examples:**

>>> f.has_data_axes()
True

>>> f.has_data_axes(key='auxiliarycoordinate2')
False

        '''
        axes = self.get_data_axes(key, default=None)
        if axes is None:
            return False

        return True
    #--- End: def
    
    def del_construct(self, key, default=ValueError()):
        '''Remove a metadata construct.

If a domain axis construct is selected for removal then it can't be
spanned by any data arrays of the field nor metadata constructs, nor
be referenced by any cell method constructs. However, a domain
ancillary construct may be removed even if it is referenced by
coordinate reference construct.

.. versionadded:: 1.7.0

.. seealso:: `get_construct`, `constructs`, `has_construct`,
             `set_construct`

:Parameters:

    key: `str`
        The construct identifier of the metadata construct to be
        removed.

        *Parameter example:*
          ``key='auxiliarycoordinate0'``

    default: optional
        Return the value of the *default* parameter if the data axes
        have not been set. If set to an `Exception` instance then it
        will be raised instead.

:Returns:

        The removed metadata construct.

**Examples:**


>>> f.del_construct('auxiliarycoordinate2')
<AuxiliaryCoordinate: latitude(111, 106) degrees_north>
>>> f.del_construct('auxiliarycoordinate2')
ValueError: Can't get remove non-existent construct
>>> f.del_construct('auxiliarycoordinate2', default=False)
False

        '''
        if key in self.domain_axes and key in self.get_data_axes(default=()):
            raise ValueError(
                "Can't remove domain axis {!r} that is spanned by the data of the field construct".format(
                    key))

        return super().del_construct(key, default=default)
    #--- End: def

    def set_data(self, data, axes=None, copy=True):
        '''Set the data of the field construct.

The units, calendar and fill value properties of the data object are
removed prior to insertion.

.. versionadded:: 1.7.0

.. seealso:: `del_data`, `get_data`, `has_data`, `set_data_axes`

:Parameters:

    data: `Data`
        The data to be inserted.

    axes: (sequence of) `str`, or `None`
        The identifiers of the domain axes spanned by the data
        array. If `None` then the data axes are not set.

        The axes may also be set afterwards with the `set_data_axes`
        method.

        *Parameter example:*
          ``axes=['domainaxis2']``
        
        *Parameter example:*
          ``axes='domainaxis2'``
        
        *Parameter example:*
          ``axes=['domainaxis2', 'domainaxis1']``
        
        *Parameter example:*
          ``axes=None``
        
    copy: `bool`, optional
        If False then do not copy the data prior to insertion. By
        default the data are copied.

:Returns:

    `None`

**Examples:**

Set the domain axis constructs spanned by the data of the field
construct:

>>> d
<Data(10, 9): [[23.6, ..., 76.8]]>
>>> f.set_data(d, axes=['domainaxis0', 'domainaxis1'])
>>> f.set_data(d)

        '''
        if axes is None:
            existing_axes = self.get_data_axes(default=None)
            if existing_axes is not None:                
                self.set_data_axes(axes=existing_axes, _shape=data.shape)
        else:
            self.set_data_axes(axes=axes, _shape=data.shape)

        super().set_data(data, copy=copy)
    #--- End: def

    def set_data_axes(self, axes, key=None, _shape=None):
        '''Set the domain axis constructs spanned by the data of the field or
of a metadata construct.

.. versionadded:: 1.7.0

.. seealso:: `del_data_axes`, `get_data`, `get_data_axes`,
             `has_data_axes`

:Parameters:

     axes: sequence of `str`
        The identifiers of the domain axis constructs spanned by the
        data of the field or of a metadata construct.

        *Parameter example:*
          ``axes='domainaxis1'``

        *Parameter example:*
          ``axes=['domainaxis1']``

        *Parameter example:*
          ``axes=['domainaxis1', 'domainaxis0']``

     key: `str`, optional
        Specify a metadata construct, instead of the field construct.

        *Parameter example:*
          ``key='domainancillary1'``

:Returns:

    `None`

**Examples:**

Set the domain axis constructs spanned by the data of the field
construct:

>>> f.set_data_axes(['domainaxis0', 'domainaxis1'])

Set the domain axis constructs spanned by the data of a metadata
construct:

>>> f.set_data_axes(['domainaxis1'], key='dimensioncoordinate1')

        '''
        if isinstance(axes, basestring):
            axes = (axes,)            
            
        if key is not None:
            return super().set_data_axes(axes=axes, key=key)

        if _shape is None:
            data = self.get_data(None)
            if data is not None:
                _shape = data.shape
        #--- End: if
            
        if _shape is not None:
            domain_axes = self.constructs.filter_by_type('domain_axis')
            axes_shape = []
            for axis in axes:
                if axis not in domain_axes:
                    raise ValueError(
                        "Can't set field construct data axes: Domain axis {!r} doesn't exist".format(
                            axis))
    
                axes_shape.append(domain_axes[axis].get_size())
    
            if _shape != tuple(axes_shape):
                raise ValueError(
                    "Can't set field construct data axes: Data array shape of {!r} does not match the shape of the given domain axes {}: {}".format(
                        _shape, tuple(axes), tuple(axes_shape)))
        #--- End: if
        
        axes = tuple(axes)
        self._set_component('data_axes', axes, copy=False)

        self.constructs._field_data_axes = axes
    #--- End: def

#--- End: class



from future.utils import with_metaclass
from builtins import super

import abc

from copy import deepcopy

from . import Container


class Parameters(with_metaclass(abc.ABCMeta, Container)):
    '''Abstract base class for a collection of named parameters.

.. versionadded:: 1.7.0
    '''
    def __init__(self, parameters=None, source=None, copy=True): 
        '''**Initialization**

:Parameters:

    parameters: `dict`, optional
       Set parameters. The dictionary keys are parameter names, with
       corresponding values. Ignored if the *source* parameter is set.

       *Parameter example:*
         ``parameters={'earth_radius': 6371007.}``

       Parameters may also be set after initialisation with the
       `set_parameters` and `set_parameter` methods.

    source: optional
        Initialize the parameters from those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(source=source, copy=copy)

        self._set_component('parameters', {}, copy=False)

        if source:
            try:
                parameters = source.parameters()
            except AttributeError:
                parameters = None
        #--- End: if
        
        if parameters is None:
            parameters = {}
            copy = False

        self.set_parameters(parameters, copy=copy)
    #--- End: def

    def clear_parameters(self):
        '''Remove all parameters.

.. versionadded:: 1.7.0

.. seealso:: `del_parameter`, `parameters`, `set_parameters`

:Returns:

    `dict`
        The parameters that have been removed.

**Examples:**

>>> old = f.clear_parameters()
>>> old
{'standard_parallel': 25.0;
 'longitude_of_central_meridian': 265.0,
 'latitude_of_projection_origin': 25.0}
>>> f.parameters()
{}
>>> f.set_parameters(old)
>>> f.parameters()
{'standard_parallel': 25.0;
 'longitude_of_central_meridian': 265.0,
 'latitude_of_projection_origin': 25.0}

        '''
        out = self._get_component('parameters')
        self._set_component('parameters', {})
        return out.copy()
    #--- End: def
    
#    def copy(self):
#        '''Return a deep copy.
#
#``f.copy()`` is equivalent to ``copy.deepcopy(f)``.
#
#.. versionadded:: 1.7.0
#
#:Returns:
#
#        The deep copy.
#
#**Examples:**
#
#>>> g = f.copy()
#
#        '''
#        return type(self)(source=self, copy=True)
#    #--- End: def
    
    def del_parameter(self, parameter, default=ValueError()):
        '''Delete a parameter.

.. versionadded:: 1.7.0

.. seealso:: `get_parameter`, `has_parameter`, `parameters`,
             `set_parameter`

:Parameters:

    parameter: `str`
        The name of the parameter to be deleted.

    default: optional
        Return the value of the *default* parameter if the parameter
        has not been set. If set to an `Exception` instance then it
        will be raised instead.

:Returns:

        The removed parameter. If the parameter has not been then
        *default* is returned, if provided.

**Examples:**

>>> f.set_parameter('earth_radius', 6371007)
>>> f.has_parameter('earth_radius')
True      
>>> f.get_parameter('earth_radius')
6371007   
>>> f.del_parameter('earth_radius')
6371007   
>>> f.has_parameter('earth_radius')
False
>>> print(f.del_parameter('earth_radius', None))
None
>>> print(f.get_paramete('earth_radius', None))
None

        '''
        try:
            return self._get_component('parameters').pop(parameter)
        except KeyError:
            return self._default(default,
                                 "{!r} has no {!r} parameter".format(
                                     self.__class__.__name__, parameter))
    #--- End: def

    def get_parameter(self, parameter, default=ValueError()):
        '''Get a parameter value.

.. versionadded:: 1.7.0

:Parameters:

    parameter: `str`
        The name of the parameter.

    default: optional
        Return the value of the *default* parameter if the parameter
        has not been set. If set to an `Exception` instance then it
        will be raised instead.

:Returns:

        The value of the parameter.

**Examples:**

>>> f.set_parameter('earth_radius', 6371007)
>>> f.has_parameter('earth_radius')
True      
>>> f.get_parameter('earth_radius')
6371007   
>>> f.del_parameter('earth_radius')
6371007   
>>> f.has_parameter('earth_radius')
False
>>> print(f.del_parameter('earth_radius', None))
None
>>> print(f.get_parameter('earth_radius', None))
None

        '''
        try:
            return self._get_component('parameters')[parameter]
        except KeyError:
            return self._default(default,
                                 "{!r} has no {!r} parameter".format(
                                     self.__class__.__name__, parameter))
    #--- End: def
    
    def has_parameter(self, parameter):
        '''Whether a parameter has been set.

.. versionadded:: 1.7.0

.. seealso:: `del_parameter`, `get_parameter`, `parameters`,
             `set_parameter`

:Parameters:

    parameter: `str`
        The name of the parameter.

        *Parameter example:*
           ``parameter='geoid_name'``

:Returns:

     `bool`
        True if the parameter has been set, otherwise False.

**Examples:**

>>> f.set_parameter('earth_radius', 6371007)
>>> f.has_parameter('earth_radius')
True      
>>> f.get_parameter('earth_radius')
6371007   
>>> f.del_parameter('earth_radius')
6371007   
>>> f.has_parameter('earth_radius')
False
>>> print(f.del_parameter('earth_radius', None))
None
>>> print(f.get_parameter('earth_radius', None))
None

        '''
        return parameter in self._get_component('parameters')
    #--- End: def

    def parameters(self):
        '''Return all parameters.

.. versionadded:: 1.7.0

.. seealso:: `clear_parameters`, `get_parameter`, `has_parameter`
             `set_parameters`

:Returns:

    `dict`
        The parameters.

**Examples:**

>>> old = f.clear_parameters()
>>> old
{'standard_parallel': 25.0;
 'longitude_of_central_meridian': 265.0,
 'latitude_of_projection_origin': 25.0}
>>> f.parameters()
{}
>>> f.set_parameters(old)
>>> f.parameters()
{'standard_parallel': 25.0;
 'longitude_of_central_meridian': 265.0,
 'latitude_of_projection_origin': 25.0}

        '''
        return self._get_component('parameters').copy()
    #--- End: def

    def set_parameters(self, parameters, copy=True):
        '''Set parameters.

.. versionadded:: 1.7.0

.. seealso:: `clear_parameters`, `parameters`, `set_parameter`

:Parameters:

    parameters: `dict` 
        Store the parameters from the dictionary supplied.

        *Parameter example:*
          ``parameters={'earth_radius': 6371007}``

    copy: `bool`, optional
        If False then any parameter values provided by the
        *parameters* parameter are not copied before insertion. By
        default they are deep copied.

:Returns:

    `None`

**Examples:**

>>> old = f.clear_parameters()
>>> old
{'standard_parallel': 25.0;
 'longitude_of_central_meridian': 265.0,
 'latitude_of_projection_origin': 25.0}
>>> f.parameters()
{}
>>> f.set_parameters(old)
>>> f.parameters()
{'standard_parallel': 25.0;
 'longitude_of_central_meridian': 265.0,
 'latitude_of_projection_origin': 25.0}

        '''
        if copy:
            parameters = deepcopy(parameters)                
        else:
            parameters = parameters.copy()
        
        self._get_component('parameters').update(parameters)
    #--- End: def

    def set_parameter(self, term, value, copy=True):
        '''Set a parameter-valued term.

.. versionadded:: 1.7.0

.. seealso:: `parameters`

:Returns:

    `None`

**Examples:**

>>> f.set_parameter('earth_radius', 6371007)
>>> f.has_parameter('earth_radius')
True      
>>> f.get_parameter('earth_radius')
6371007   
>>> f.del_parameter('earth_radius')
6371007   
>>> f.has_parameter('earth_radius')
False
>>> print(f.del_parameter('earth_radius', None))
None
>>> print(f.get_parameter('earth_radius', None))
None

     '''
        if copy:
            value = deepcopy(value)
            
        self._get_component('parameters')[term] = value
    #--- End: def
    
#--- End: class



from future.utils import with_metaclass
from builtins import (range, super)

import abc

import numpy

from .array import Array


class CompressedArray(with_metaclass(abc.ABCMeta, Array)):
    '''Abstract base class for a container of an underlying compressed
array.

See `cfdm.GatheredArray` for an example implementation.

.. versionadded:: 1.7.0

    '''
    def __init__(self, compressed_array=None, shape=None, size=None,
                 ndim=None, compressed_dimension=None,
                 compression_type=None, **kwargs):
        '''**Initialization**

:Parameters:

    compressed_array: subclass of `Array`
        The compressed array.

    shape: `tuple`
        The uncompressed array dimension sizes.

    size: `int`
        Number of elements in the uncompressed array.

    ndim: `int`
        The number of uncompressed array dimensions

    compressed_dimension: `int`
        The position of the compressed dimension in the compressed
        array.

    compression_type: `str`
        The type of compression.        
        
    kwargs: *optional*
        Further named parameters and their values needed to define the
        compressed array.

        '''
        super().__init__(shape=shape, size=size, ndim=ndim,
                         compressed_dimension=compressed_dimension,
                         compression_type=compression_type, **kwargs)

        self._set_compressed_Array(compressed_array, copy=False)
    #--- End: def

    @abc.abstractmethod
    def __getitem__(self, indices):
        '''Return an uncompressed subspace as an independent numpy array.

x.__getitem__(indices) <==> x[indices]

The indices that define the subspace are relative to the uncompressed
array.

Indexing follows rules that are very similar to the numpy indexing
rules, the only differences being:

* An integer index i takes the i-th element but does not reduce the
  rank by one.

..

* When two or more dimensions' indices are sequences of integers then
  these indices work independently along each dimension (similar to
  the way vector subscripts work in Fortran). This is the same
  behaviour as indexing on a Variable object of the netCDF4 package.

.. versionadded:: 1.7.0

        '''
        raise NotImplementedError() # pragma: no cover
    #--- End: def

    def _get_compressed_Array(self, default=ValueError()):
        '''TODO

:Parameters:

    default: optional
        Return the value of the *default* parameter if the array
        has not been set. If set to an `Exception` instance then it
        will be raised instead.

:Returns:

        The compressed Array instance.

**Examples:**

>>> a = d.get_data(None)

        '''
        return self._get_component('compressed_Array', default)
    #--- End: def

    def _set_compressed_Array(self, array, copy=True):
        '''Set the compressed array.

.. versionadded:: 1.7.0

:Parameters:

    array: subclass of `cfdm.data.Array`
        The compressed data to be inserted.

:Returns:

    `None`

**Examples:**

>>> d._set_compressed_Array(a)

        '''
#        if not isinstance(array, Array):
#            raise TypeError("asdads  0000000000000000000 TODO")

        if copy:
            array = array.copy()

        self._set_component('compressed_Array', array, copy=False)
    #--- End: def

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def array(self):
        '''Return an independent numpy array containing the uncompressed data.

:Returns:

    `numpy.ndarray`
        The uncompressed array.

**Examples:**

>>> n = a.array
>>> isinstance(n, numpy.ndarray)
True

        '''
        return self[...]
    #--- End: def

    @property
    def dtype(self):
        '''Data-type of the data elements.

**Examples:**

>>> a.dtype
dtype('float64')
>>> print(type(a.dtype))
<type 'numpy.dtype'>

        '''
        return self._get_compressed_Array().dtype
    #--- End: def

    @property
    def ndim(self):
        '''The number of dimensions of the uncompressed data.

**Examples:**

>>> d.shape
(73, 96)
>>> d.ndim
2
>>> d.size
7008

>>> d.shape
(1, 1, 1)
>>> d.ndim
3
>>> d.size
1

>>> d.shape
()
>>> d.ndim
0
>>> d.size
1

        '''
        return self._get_component('ndim')
    #--- End: def

    @property
    def shape(self):
        '''Shape of the uncompressed data.

**Examples:**

>>> d.shape
(73, 96)
>>> d.ndim
2
>>> d.size
7008

>>> d.shape
(1, 1, 1)
>>> d.ndim
3
>>> d.size
1

>>> d.shape
()
>>> d.ndim
0
>>> d.size
1

        '''
        return self._get_component('shape')
    #--- End: def

    @property
    def size(self):
        '''Number of elements in the uncompressed data.

**Examples:**

>>> d.shape
(73, 96)
>>> d.size
7008
>>> d.ndim
2

>>> d.shape
(1, 1, 1)
>>> d.ndim
3
>>> d.size
1

>>> d.shape
()
>>> d.ndim
0
>>> d.size
1

        '''
        return self._get_component('size')
    #--- End: def

    @property
    def compressed_array(self):
        '''Return an independent numpy array containing the compressed data.

:Returns:

    `numpy.ndarray`
        The compressed array.

**Examples:**

>>> n = a.compressed_array
>>> import numpy
>>> isinstance(n, numpy.ndarray)
True

        '''
        ca = self._get_compressed_Array(None)
        if ca is None:
            raise ValueError("There is no underlying compressed array")
        
        return ca.array
    #--- End: def

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def get_compressed_axes(self):
        '''Return axes that are compressed in the underlying array.

:Returns:

    `list`
        The compressed axes described by their integer positions.

**Examples:**

>>> c.ndim
4
>>> c.compressed_array.ndim
3
>>> c.compressed_axes()
[1, 2]

        '''
        compressed_dimension = self.get_compressed_dimension()
        compressed_ndim = self._get_compressed_Array().ndim
        
        return list(range(compressed_dimension, self.ndim - (compressed_ndim - compressed_dimension - 1)))
    #--- End: def

    def get_compressed_dimension(self, *default):
        '''Return the position of the compressed dimension in the compressed
array.

.. versionadded:: 1.7.0

.. seealso:: `get_compressed_axearray`, `get_compressed_axes`,
             `get_compressed_type`

:Parameters:

    default: optional
        Return *default* if the underlying array is not compressed.

:Returns:

    `int`
        The position of the compressed dimension in the compressed
        array. If the underlying is not compressed then *default* is
        returned, if provided.

**Examples:**

>>> i = d.get_compressed_dimension()

        '''
        return self._get_component('compressed_dimension', *default)
    #--- End: def

    @abc.abstractmethod
    def to_memory(self):
        '''TODO

:Returns:
        TODO
  
**Examples:**
     TODO
 '''
        self._set_compressed_Array(self._get_compressed_Array().to_memory())
        return self
    #--- End def
    
    def source(self, default=ValueError()):
        '''TODO Return the underlying array object.

:Parameters:

    default: optional
        Return the value of the *default* parameter if the array
        has not been set. If set to an `Exception` instance then it
        will be raised instead.

:Returns:

    subclass of `Array`
        TODO The underlying array object.

**Examples:**

>>> TODO

        '''
        return self._get_compressed_Array(default=default)
    #--- End: def
        
#--- End: class



from .abstract           import IO, IORead, IOWrite, Implementation
from .cfdmimplementation import (CFDMImplementation,
                                 _implementation,
                                 implementation)
from .read               import read
from .write              import write



from builtins import object


class RaggedContiguous(object):
    '''Mixin class for an underlying compressed ragged array.

.. versionadded:: 1.7.0

    '''
    def get_count(self, default=ValueError()):
        '''Return the countcount_va variable for a compressed array.

.. versionadded:: 1.7.0

:Parameters:

    default: optional
        Return the value of the *default* parameter if a count
        variable has not been set. If set to an `Exception` instance
        then it will be raised instead.

:Returns:

        The count variable.

**Examples:**

>>> c = d.get_count()

        '''
        try:
            return self._get_component('count_variable')
        except ValueError:
            return self._default(default,
                                 "{!r} has no count variable".format(
                                     self.__class__.__name__))
    #--- End: def

#--- End: class


class RaggedIndexed(object):
    '''Mixin class for an underlying indexed ragged array.

    '''
    def get_index(self, default=ValueError()):
        '''Return the index variable for a compressed array.

.. versionadded:: 1.7.0

:Parameters:

    default: optional
        Return the value of the *default* parameter if an index
        variable has not been set. If set to an `Exception` instance
        then it will be raised instead.

:Returns:

        The index variable.

**Examples:**

>>> i = d.get_index()

        '''
        try:
            return self._get_component('index_variable')
        except ValueError:
            return self._default(default,
                                 "{!r} has no index variable".format(
                                     self.__class__.__name__))
    #--- End: def
    
#--- End: class
